#--
# mpi.py
# --
# this module contains hooks for parallel computing with the standard Message Passing Interface (MPI).
#
# variables and functions
#     * rank       = index of this node.
#     * size       = total number of nodes.
#     * barrier()  = halt execution until all nodes have called barrier().
#     * finalize() = terminate the MPI session (otherwise if one node finishes before the others they may be killed as well).
#
from __future__ import absolute_import, print_function
from pspy import so_misc
import warnings
import sys
import numpy as np

# blank function template
def _pass():
    pass

# set the default value
_initialized = False
_switch      = False
rank    = 0
size    = 1
comm    = None
barrier = _pass
broadcast = _pass

def init(switch = False):
    ''' initialize MPI set-up '''
    global _initialized, _switch
    global rank, size, comm
    global barrier, finalize

    exit_code = 0

    if isinstance(switch, str):
        switch = so_misc.str2bool(switch)
    else:
        pass

    if not _initialized:
        _initialized = True
    else:
        print("MPI is already intialized")
        return exit_code 

    if not switch:
        print("WARNING: MPI is turned off by default. Use mpi.init(switch=True) to initialize MPI")
        print("MPI is turned off")
        return exit_code
    else:
        _switch = True

    try:        
        from  mpi4py import MPI
        comm    = MPI.COMM_WORLD
        rank    = comm.Get_rank()
        size    = comm.Get_size()
        barrier = comm.Barrier
        print("MPI: rank %d is initalized" %rank)

    except ImportError as exc:
        sys.stderr.write("IMPORT ERROR: " + __file__ + " (" + str(exc) + "). Could not load mpi4py. MPI will not be used.\n")

def is_initialized():
    global _initialized
    return _initialized

def is_mpion():
    global _switch
    return _switch

def taskrange(imax, imin = 0, shift = 0):
    '''
        Given the range of tasks, it returns the subrange of the tasks which the current processor have to perform.
        
            -- input
            -ntask: the number of tasks
            -shift: manual shift to the range
            
            -- output
            - range
        
        For size 2, rank 1
        e.g) taskrange(0, 9, offset = 0) -> [5,6,7,8,9]
        e.g) taskrange(0, 9, offset = 2) -> [7,8,9,10,11]

    '''
    global rank, size

    if not isinstance(imin, int) or not isinstance(imax, int) or not isinstance(shift, int):
        raise TypeError("imin, imax and shift must be integers")
    elif not is_initialized():
        print("MPI is not yet properly initialized. Are you sure this is what you want to do?")
    else:
        pass

    ntask = imax - imin + 1

    subrange = None
    if ntask <= 0 :
        print("number of task can't be zero")
        subrange = np.arange(0,0) # return zero range
    else:
        if size > ntask:
            delta     = 1
            remainder = 0
        else:
            delta     = ntask // size
            remainder = ntask % size

        # correction for remainder 
        start      = imin + rank*delta
        scorr      = min(rank, remainder)
        start      += scorr

        delta      += 1 if rank < remainder else 0

        end        = start + delta
        end        = min(end, imax+1)

        print("rank: %d size: %d ntask: %d start: %d end: %d" %(rank, size, ntask, start, end-1))

        subrange   = np.arange(start, end)

    return subrange


def transfer_data(data, tag, dest=0, mode='append'):
    if not is_initialized():
        raise ValueError("mpi is yet initalized")
    elif not is_mpion():
        raise ValueError("mpi is not on")
    else:
        pass

    global rank, size, comm

    senders = range(0,size)
    senders.pop(dest)

    if rank != dest:
        print("rank%d is sending data to rank%d with tag%d" %(rank, dest, tag))
        comm.send(data, dest, tag=tag)
    else:
        for sender in senders:
            print("rank%d is receiving tag%d data from rank%d" %(dest,tag,sender))
            if type(data) == dict:
                recv_data = comm.recv(source=sender, tag=tag)
                so_misc.merge_dict(data,recv_data)
            elif type(data) == np.ndarray:
                recv_data = comm.recv(source=sender, tag=tag)
                if mode == 'append':
                    data = np.append(data,recv_data)
                elif mode == 'add':
                    data += recv_data
                else:
                    assert(0)
            else:
                raise NotImplementedError()

    # wait for the end of the data transfer
    comm.Barrier()
    return data

def broadcast(data, root=0):
    comm.bcast(data, root=root)
    comm.Barrier()
























    






"""
@brief: python routines for reading a dictionnary file
"""


def ask_for(key):
    s = input("so_dict: enter value for '%s': " % key)
    try:
        val = eval(s)
    except NameError:
        # allow people to enter unquoted strings
        val = s
    return val


class so_dict(dict):
    def __init__(self, ask=False):
        """
        @param ask if the dict doesn't have an entry for a key, ask for the associated value and assign
        """
        dict.__init__(self)
        self.ask = ask

    def __getitem__(self, key):
        if key not in self:
            if self.ask:
                print("so_dict: parameter '%s' not found" % key)
                val = ask_for(key)
                print("so_dict: setting '%s' = %s" % (key, repr(val)))
                dict.__setitem__(self, key, val)
            else:
                raise ValueError("Missing '{}' key value in dictionary".format(key))
        return dict.__getitem__(self, key)

    def get(self, key, default=None, raise_error=False):
        res = super().get(key, default)
        if res is None and raise_error:
            raise ValueError("Missing '{}' key value in dictionary".format(key))
        return res

    def read_from_file(self, filename):
        f = open(filename)
        old = ""
        for line in f:
            line = line.strip()
            if len(line) == 0 or line[0] == "#":
                continue
            s = line.split("#")
            line = s[0]
            s = line.split("\\")
            if len(s) > 1:
                old = old.join(s[0])
                continue
            else:
                line = old.join(s[0])
                old = ""
            for i in range(len(line)):
                if line[i] != " ":
                    line = line[i:]
                    break
            exec(line, globals())
            s = line.split("=")
            if len(s) != 2:
                print("Error parsing line:")
                print(line)
                continue
            key = s[0].strip()
            val = eval(s[1].strip())  # XXX:make safer
            self[key] = val
        f.close()

    readFromFile = read_from_file

    def write_to_file(self, filename, mode="w"):
        f = open(filename, mode)
        keys = self.keys()
        keys.sort()
        for key in keys:
            f.write("%s = %s\n" % (key, repr(self[key])))
        f.close()

    writeToFile = write_to_file



"""Tests pspy versus namaster."""
import os
import tempfile
import unittest

import healpy as hp
import numpy as np
from pspy import pspy_utils, so_map, so_mcm, so_spectra, so_window, sph_tools


def _assert_import(pkg_name):
    try:
        import importlib

        importlib.import_module(pkg_name)
    except ModuleNotFoundError:
        raise ModuleNotFoundError(f"You need to install '{pkg_name}' to use this unit test")


_assert_import("camb")
_assert_import("pymaster")

import camb
import pymaster as nmt

output_dir = os.path.join(tempfile.gettempdir(), "test_pspy_namaster")
os.makedirs(output_dir, exist_ok=True)

lmin, lmax = 2, 10**4
l = np.arange(lmin, lmax)
cosmo_params = {
    "H0": 67.5,
    "As": 1e-10 * np.exp(3.044),
    "ombh2": 0.02237,
    "omch2": 0.1200,
    "ns": 0.9649,
    "Alens": 1.0,
    "tau": 0.0544,
}
pars = camb.set_params(**cosmo_params)
pars.set_for_lmax(lmax, lens_potential_accuracy=1)
results = camb.get_results(pars)
powers = results.get_cmb_power_spectra(pars, CMB_unit="muK")
cl_file = os.path.join(output_dir, "cl_camb.dat")
np.savetxt(cl_file, np.hstack([l[:, np.newaxis], powers["total"][lmin:lmax]]))

lon, lat = 30, 50
radius = 25
nside = 512
lmax = 3 * nside - 1
ncomp = 3
niter = 3
spectra = ["TT", "TE", "TB", "ET", "BT", "EE", "EB", "BE", "BB"]

template_healpix = so_map.healpix_template(ncomp, nside=nside)

binary_healpix = so_map.healpix_template(ncomp=1, nside=nside)
vec = hp.ang2vec(lon, lat, lonlat=True)
disc = hp.query_disc(nside, vec, radius=np.deg2rad(radius))
binary_healpix.data[disc] = 1

cmb = template_healpix.synfast(cl_file)
noise = so_map.white_noise(cmb, rms_uKarcmin_T=20, rms_uKarcmin_pol=np.sqrt(2) * 20)
cmb.data += noise.data

window = so_window.create_apodization(binary_healpix, apo_type="C1", apo_radius_degree=1)
mask = so_map.simulate_source_mask(binary_healpix, n_holes=10, hole_radius_arcmin=30)
mask = so_window.create_apodization(mask, apo_type="C1", apo_radius_degree=1)
window.data *= mask.data
window = (window, window)


def run_pspy():
    binning_file = os.path.join(output_dir, "binning.dat")
    pspy_utils.create_binning_file(bin_size=40, n_bins=100, file_name=binning_file)
    mbb_inv, Bbl = so_mcm.mcm_and_bbl_spin0and2(
        window, binning_file, lmax=lmax, type="Cl", niter=niter
    )
    alms = sph_tools.get_alms(cmb, window, niter=niter, lmax=lmax)
    ell, ps = so_spectra.get_spectra(alms, spectra=spectra)
    lb, Clb = so_spectra.bin_spectra(
        ell, ps, binning_file, lmax, type="Cl", mbb_inv=mbb_inv, spectra=spectra
    )
    return lb, Clb


def run_namaster():
    field_0 = nmt.NmtField(window[0].data, [cmb.data[0]])
    field_2 = nmt.NmtField(window[1].data, [cmb.data[1], cmb.data[2]])
    nlb = 40
    b = nmt.NmtBin(nside, nlb=nlb)
    lb = b.get_effective_ells()
    wsp = nmt.NmtWorkspace()
    wsp.compute_coupling_matrix(field_0, field_2, b, is_teb=True, n_iter=niter, lmax_mask=lmax)
    cl_coupled_00 = nmt.compute_coupled_cell(field_0, field_0)
    cl_coupled_02 = nmt.compute_coupled_cell(field_0, field_2)
    cl_coupled_22 = nmt.compute_coupled_cell(field_2, field_2)
    cls_coupled = np.array(
        [
            cl_coupled_00[0],  # TT
            cl_coupled_02[0],  # TE
            cl_coupled_02[1],  # TB
            cl_coupled_22[0],  # EE
            cl_coupled_22[1],  # EB
            cl_coupled_22[2],  # BE
            cl_coupled_22[3],  # BB
        ]
    )
    cls_uncoupled = wsp.decouple_cell(cls_coupled)
    Clb = {k: cls_uncoupled[i] for i, k in enumerate(["TT", "TE", "TB", "EE", "EB", "BE", "BB"])}
    Clb["ET"] = Clb["TE"]
    Clb["BT"] = Clb["TB"]
    return lb, Clb


class SOPspyNamasterTests(unittest.TestCase):
    def setUp(self):
        pass

    def test_pspy_namaster(self):
        lb_pspy, Clb_pspy = run_pspy()
        lb_nmt, Clb_nmt = run_namaster()

        decimal = 12
        for spec in spectra:
            msg = f"Testing {spec} spectrum"
            np.testing.assert_almost_equal(
                Clb_pspy[spec], Clb_nmt[spec], err_msg=msg, decimal=decimal
            )


if __name__ == "__main__":
    unittest.main()



"""
@brief: misc python routines 
"""
from __future__ import absolute_import, print_function

def is_empty(s):
    '''
        if s is either '' or None, this function returns True. Otherwise, false.
    '''
    return True if (s is None) or (len(s) == 0) else False

def get_from_dict(nested_dict, keys, safe=True):
    if not type(keys) is tuple: keys = (keys,)
    if safe and not has_key(nested_dict, keys): return None

    if(len(keys) > 1):
        return get_from_nested_dict(nested_dict[keys[0]], keys[1:], False)
    else:
        return nested_dict[keys[0]]

def has_key(nested_dict, keys):
    ''' search through nested dictionary to fine the elements '''
    if not type(keys) is tuple: keys = (keys,)
    if not type(nested_dict) == dict: return False

    if(len(keys) > 1):
        has_it = nested_dict.has_key(keys[0])
        return has_key(nested_dict[keys[0]], keys[1:]) if has_it else False
    else:
        return nested_dict.has_key(keys[0])

def str2bool(v):
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise TypeException("Can't convert 'str' object to 'boolean'")

def merge_dict(a, b, path=None, clean = True):
    '''
    merges b into a

    ref: https://stackoverflow.com/questions/7204805/dictionaries-of-dictionaries-merge"
    '''
    if path is None: path = []

    for key in b:
        if key in a:
            if isinstance(a[key], dict):
                if isinstance(b[key], dict):
                    merge_dict(a[key], b[key], path + [str(key)])
                else:
                    a[key] = b[key]
            elif isinstance(a[key], np.ndarray) and np.array_equal(a[key], b[key]):
                pass
            elif a[key] == b[key]:
                pass # same leaf value
            else:
                raise Exception('Conflict at %s' % '.'.join(path + [str(key)]))
        else:
            a[key] = b[key]

    if clean: del b

    return a




"""
Routines for handling flat maps (inspired by flipper from sudeep das)
"""
import numpy as np
from pixell import enmap
from pspy import so_map, pspy_utils
import pylab as plt
from copy import deepcopy

class fft2D:
    """
    class describing the two dimensional FFTs of a so_map
    """
    def __init__(self):
        pass

    def copy(self):
        return copy.deepcopy(self)
        
    def trim_fft(self, lmax):
    
        ft = fft2D()
        ft.ncomp = self.ncomp

        ly, lx = self.ly, self.lx
        idy = np.where((ly < lmax) & (ly > -lmax))[0]
        idx = np.where((lx < lmax) & (lx > -lmax))[0]
        npix_y, npix_x = len(idy), len(idx)
    
        if self.ncomp == 1:
            trimmed_fft = np.zeros((npix_y, npix_x), dtype=complex)
            trim = self.kmap[idy, :]
            trimmed_fft = trim[:, idx]
        else:
            trimmed_fft = np.zeros((3, npix_y, npix_x), dtype=complex)
            for i in range(3):
                trim = self.kmap[i, idy, :]
                trimmed_fft[i] = trim[:, idx]

        ft.kmap = trimmed_fft
        
        lxmap = self.lxmap[idy,: ]
        ft.lxmap = lxmap[:,idx]
        
        lymap = self.lymap[idy,:]
        ft.lymap = lymap[:,idx]
        
        ft.ly, ft.lx = ft.lymap[:,0], ft.lxmap[0,:]
        
        ft.lmap = np.sqrt(ft.lxmap**2 + ft.lymap**2)
        ft.thetamap = np.arctan2(ft.lymap,ft.lxmap) * 180/np.pi

        return ft
        
        
    def create_kspace_mask(self, vertical_stripe=None, horizontal_stripe=None):
        mask = np.real(self.kmap.copy())
        if mask.ndim == 3:
            mask = mask[0]
        mask[:,:] = 1.
        if vertical_stripe is not None:
            idx = np.where((self.lx < vertical_stripe[1]) & (self.lx > vertical_stripe[0]))
            mask[:,idx] = 0.
        if vertical_stripe is not None:
            idy = np.where((self.ly < horizontal_stripe[1]) & (self.ly > horizontal_stripe[0]))
            mask[idy,:] = 0.
        self.kmask = mask
        
        
    def map_from_fft(self, normalize="phys"):
        return enmap.ifft(self.kmap, normalize=normalize)

        
class power2D:
    """
    a class describing the 2-D power spectra of a so_map
    """
    def __init__(self):
        pass
        
    def copy(self):
        return deepcopy(self)

    def create_kspace_mask(self, vertical_stripe=None, horizontal_stripe=None):
        mask = self.powermap["II"].copy()
        mask[:,:] = 1.
        if vertical_stripe is not None:
            idx = np.where((self.lx < vertical_stripe[1]) & (self.lx > vertical_stripe[0]))
            mask[:,idx] = 0.
        if vertical_stripe is not None:
            idy = np.where((self.ly < horizontal_stripe[1]) & (self.ly > horizontal_stripe[0]))
            mask[idy,:] = 0.
        self.kmask = mask

    def plot(self,
             log=False,
             title="",
             power_of_ell=0,
             show_bins_from_files=None,
             draw_circles_at_ell=None,
             draw_vertical_lines_at_ell=None,
             value_range=None,
             show= True,
             png_file=None):
             


        for spec in self.spectra:
            p =  self.powermap[spec].copy()
     
            p[:] *= (self.lmap + 1.)**power_of_ell
            p = np.fft.fftshift(p)
     
            if show_bins_from_files is not None:
                bin_lo, bin_hi, bin_c, bin_size = pspy_utils.read_binning_file(binning_file, lmax)
                theta = np.arange(0, 2. * np.pi + 0.05, 0.05)
         
                for i in xrange(len(bin_lo)):
                    x, y = bin_hi[i] * np.cos(theta), bin_hi[i] * np.sin(theta)
                    plt.plot(x,y,'k')

            if draw_circles_at_ell is not None:
                for ell in draw_circles_at_ell:
                    theta = np.arange(0, 2. * np.pi + 0.05, 0.05)
                    x, y = ell * np.cos(theta), ell * np.sin(theta)
                    plt.plot(x, y, "k")
                    if len(draw_circles_at_ell) < 5:
                        plt.text(ell * np.cos(np.pi / 4.),
                                   ell * np.sin(np.pi / 4.),
                                   "%d" % np.int(ell),
                                   rotation=-45,
                                   horizontalalignment="center",
                                   verticalalignment="bottom",
                                   fontsize=8)
                 
            if draw_vertical_lines_at_ell is not None:
                for ell in draw_vertical_lines_at_ell:
                    plt.axvline(ell)
             
            if log:
                p = np.log10(np.abs(p))

            vmin = p.min()
            vmax = p.max()

            if value_range is not None:
                vmin = value_range[0]
                vmax = value_range[1]
            im = plt.imshow(p,
                            origin="down",
                            extent=[np.min(self.lx), np.max(self.lx), np.min(self.ly), np.max(self.ly)],
                            aspect="equal",
                            vmin=vmin,
                            vmax=vmax)
            plt.title(title+ "%s"%spec, fontsize=13)
            plt.colorbar()
            plt.xlabel(r'$\ell_x$',fontsize=15)
            plt.ylabel(r'$\ell_y$',fontsize=15)
     
            if png_file is not None:
                plt.savefig("%s_%s.png" % (png_file, spec))
                plt.clf()
                plt.close()
            else:
                plt.show()



def fft_from_so_map(so_map, normalize="phys"):
    """
    Creates an fft2D object out of a so_map
    """
    ft = fft2D()

    ft.lymap, ft.lxmap = so_map.data.lmap()
    ft.ly, ft.lx = ft.lymap[:,0], ft.lxmap[0,:]
    ft.lmap = np.sqrt(ft.lxmap**2 + ft.lymap**2)
    ft.thetamap = np.arctan2(ft.lymap, ft.lxmap) * 180 / np.pi
    ft.ncomp = so_map.ncomp
    ft.kmap = enmap.fft(so_map.data, normalize=normalize)
    
    return ft
    
def power_from_fft(ft, ft2=None, type = "Cl"):
    """
    Creates an 2-D power spectra object out of fft2D objects
    """

    p2d = power2D()
    p2d.lymap, p2d.lxmap = ft.lymap, ft.lxmap
    p2d.ly, p2d.lx = ft.ly, ft.lx
    p2d.lmap = ft.lmap
    p2d.thetamap = ft.thetamap
    p2d.powermap = {}
    
    fac = 1
    if type == "Dl":
        fac = p2d.lmap**2/(2*np.pi)
    
    if ft.ncomp == 1:
        p2d.spectra = ["II"]

        if ft2 is None:
            p2d.powermap["II"] = (ft.kmap*np.conj(ft.kmap)).real * fac
        else:
            p2d.powermap["II"]= (ft.kmap*np.conj(ft2.kmap)).real * fac
                  
    elif ft.ncomp == 3:
        spectra_list = []
        for i, m1 in enumerate(["I","Q","U"]):
            for j, m2 in enumerate(["I","Q","U"]):
                if ft2 is None:
                    p2d.powermap[m1+m2]= (ft.kmap[i]*np.conj(ft.kmap[j])).real * fac
                else:
                    p2d.powermap[m1+m2]= (ft.kmap[i]*np.conj(ft2.kmap[j])).real * fac
                spectra_list += [m1+m2]
                
        p2d.spectra = spectra_list
    
        
    return p2d.lmap, p2d


def get_ffts(so_map, window, lmax=None, normalize="phys"):

    windowed_map = so_map.copy()
    if so_map.ncomp == 3:
        windowed_map.data[0] = so_map.data[0]*window[0].data
        windowed_map.data[1] = so_map.data[1]*window[1].data
        windowed_map.data[2] = so_map.data[2]*window[1].data
    if so_map.ncomp == 1:
        windowed_map.data = so_map.data * window.data
    ffts = fft_from_so_map(windowed_map, normalize=normalize)
    if lmax is not None:
        ffts=ffts.trim_fft(lmax)
    return ffts



"""
routines for window function generation
"""

import healpy as hp
import numpy as np

from pixell import curvedsky, enmap
from pspy import sph_tools


def get_distance(binary, rmax=None):
    """Get the distance to the closest masked pixels for CAR and healpix so_map binary.

    Parameters
    ----------
    binary: ``so_map``
    a ``so_map`` with binary data (1 is observed, 0 is masked)
    """

    dist = binary.copy()
    # Make sure distances are floating point number
    dist.data = dist.data.astype(float)

    if binary.pixel == "HEALPIX":
        dist.data[:] = enmap.distance_transform_healpix(binary.data, method="heap", rmax=rmax)
        dist.data[:] *= 180 / np.pi
    if binary.pixel == "CAR":
        dist.data[:] = enmap.distance_transform(binary.data, rmax=rmax)
        dist.data[:] *= 180 / np.pi

    return dist


def create_apodization(binary, apo_type, apo_radius_degree, use_rmax=False):
    """Create a apodized window function from a binary mask.

    Parameters
    ----------
    binary: ``so_map``
      a ``so_map`` with binary data (1 is observed, 0 is masked)
    apo_type: string
      the type of apodisation you want to use ("C1","C2" or "Rectangle")
    apo_radius: float
      the apodisation radius in degrees
    """

    if use_rmax:
        rmax = (apo_radius_degree * 1.1) * np.pi / 180
    else:
        rmax = None

    if apo_type == "C1":
        window = apod_C1(binary, apo_radius_degree, rmax)
    if apo_type == "C2":
        window = apod_C2(binary, apo_radius_degree, rmax)
    if apo_type == "Rectangle":
        if binary.pixel == "HEALPIX":
            raise ValueError("No rectangle apodization for HEALPIX map")
        if binary.pixel == "CAR":
            window = apod_rectangle(binary, apo_radius_degree)

    return window


def apod_C2(binary, radius, rmax=None):
    """Create a C2 apodisation as defined in https://arxiv.org/pdf/0903.2350.pdf

    Parameters
    ----------
    binary: ``so_map``
        a ``so_map`` with binary data (1 is observed, 0 is masked)
    apo_radius: float
        the apodisation radius in degrees

    """

    if radius == 0:
        return binary
    else:
        dist = get_distance(binary, rmax)
        win = binary.copy()
        idx = np.where(dist.data > radius)
        win.data = dist.data / radius - np.sin(2 * np.pi * dist.data / radius) / (2 * np.pi)
        win.data[idx] = 1

    return win


def apod_C1(binary, radius, rmax=None):
    """Create a C1 apodisation as defined in https://arxiv.org/pdf/0903.2350.pdf

    Parameters
    ----------
    binary: ``so_map``
      a so_map with binary data (1 is observed, 0 is masked)
    apo_radius: float
      the apodisation radius in degrees

    """

    if radius == 0:
        return binary
    else:
        dist = get_distance(binary, rmax)
        win = binary.copy()
        idx = np.where(dist.data > radius)
        win.data = 1.0 / 2 - 1.0 / 2 * np.cos(-np.pi * dist.data / radius)
        win.data[idx] = 1

    return win


def apod_rectangle(binary, radius):
    """Create an apodisation for rectangle window (in CAR) (smoother at the corner)

    Parameters
    ----------
    binary: ``so_map``
      a so_map with binary data (1 is observed, 0 is masked)
    apo_radius: float
      the apodisation radius in degrees

    """

    # TODO: clean this one

    if radius == 0:
        return binary
    else:
        shape = binary.data.shape
        wcs = binary.data.wcs
        Ny, Nx = shape
        pix_scale_y, pix_scale_x = enmap.pixshape(shape, wcs)
        win = binary.copy()
        win.data = win.data * 0 + 1
        win_x = win.copy()
        win_y = win.copy()
        ones = np.ones((Ny, Nx))
        deg_to_pix_x = np.pi / 180 / pix_scale_x
        deg_to_pix_y = np.pi / 180 / pix_scale_y
        lenApod_x = int(radius * deg_to_pix_x)
        lenApod_y = int(radius * deg_to_pix_y)

        for i in range(lenApod_x):
            r = float(i)
            win_x.data[:, i] = 1.0 / 2 * (ones[:, i] - np.cos(-np.pi * r / lenApod_x))
            win_x.data[:, Nx - i - 1] = win_x.data[:, i]
        for j in range(lenApod_y):
            r = float(j)
            win_y.data[j, :] = 1.0 / 2 * (ones[j, :] - np.cos(-np.pi * r / lenApod_y))
            win_y.data[Ny - j - 1, :] = win_y.data[j, :]

        win.data = win_x.data * win_y.data

        return win


def get_spinned_windows(w, lmax, niter):
    """Compute the spinned window functions (for pure B modes method)

    Parameters
    ----------
    w: ``so_map``
      map of the window function
    lmax: integer
      maximum value of the multipole for the harmonic transform
    niter: integer
      number of iteration for the harmonic transform

    """

    template = np.array([w.data.copy(), w.data.copy()])
    w1_plus, w1_minus, w2_plus, w2_minus = w.copy(), w.copy(), w.copy(), w.copy()

    if w.pixel == "CAR":
        template = enmap.samewcs(template, w.data)

    wlm = sph_tools.map2alm(w, lmax=lmax, niter=niter)
    ell = np.arange(lmax + 1)
    filter_1 = -np.sqrt((ell + 1) * ell)
    filter_2 = np.sqrt((ell + 2) * (ell + 1) * ell * (ell - 1))

    filter_2[:1] = 0

    wlm1_e = hp.almxfl(wlm, filter_1)
    wlm2_e = hp.almxfl(wlm, filter_2)
    wlm1_b = np.zeros_like(wlm1_e)
    wlm2_b = np.zeros_like(wlm2_e)

    w1 = template.copy()
    w2 = template.copy()

    if w.pixel == "HEALPIX":
        curvedsky.alm2map_healpix(np.array([wlm1_e, wlm1_b]), w1, spin=1)
        curvedsky.alm2map_healpix(np.array([wlm2_e, wlm2_b]), w2, spin=2)
    if w.pixel == "CAR":
        curvedsky.alm2map(np.array([wlm1_e, wlm1_b]), w1, spin=1)
        curvedsky.alm2map(np.array([wlm2_e, wlm2_b]), w2, spin=2)

    w1_plus.data = w1[0]
    w1_minus.data = w1[1]
    w2_plus.data = w2[0]
    w2_minus.data = w2[1]

    return w1_plus, w1_minus, w2_plus, w2_minus



"""
Routines for generalized map2alm and alm2map (healpix and CAR).
"""
import healpy as hp
import numpy as np

from pixell import curvedsky, enmap
from pspy import so_window


def map2alm(map, niter, lmax, theta_range=None, dtype=np.complex128, tweak=True):
    """Map2alm transform (for healpix or CAR).

    Parameters
    ----------
    map: ``so_map``
      the map from which to compute the alm
    niter: integer
      the number of iteration performed while computing the alm
      not that for CAR niter=0 should be enough
    lmax:  integer
      the maximum multipole of the transform
    theta_range: list of 2 elements
      [theta_min,theta_max] in radian.
      for healpix pixellisation all pixel outside this range will be assumed to be zero.
    """
    if map.pixel == "HEALPIX":
        if theta_range is None:
            alm = hp.sphtfunc.map2alm(map.data, lmax=lmax, iter=niter)

        else:
            nside = hp.pixelfunc.get_nside(map.data)
            alm = curvedsky.map2alm_healpix(map.data,
                                            lmax=lmax,
                                            theta_min=theta_range[0],
                                            theta_max=theta_range[1])
            if niter != 0:
                map_copy = map.copy()
                for _ in range(niter):
                    alm += curvedsky.map2alm_healpix(map.data-curvedsky.alm2map_healpix(alm,map_copy.data),
                                                     lmax=lmax,
                                                     theta_min=theta_range[0],
                                                     theta_max=theta_range[1])

    elif map.pixel=="CAR":
        alm = curvedsky.map2alm(map.data, lmax=lmax, tweak=tweak)
        if niter != 0:
            map_copy = map.copy()
            for _ in range(niter):
                alm += curvedsky.map2alm(map.data-curvedsky.alm2map(alm, map_copy.data), lmax=lmax, tweak=tweak)
    else:
        raise ValueError("Map is neither a CAR nor a HEALPIX")

    alm = alm.astype(dtype)
    return alm

def alm2map(alms, so_map, tweak=True):
    """alm2map transform (for healpix and CAR).

    Parameters
    ----------
    alms: array
      a set of alms, the shape of alms should correspond to so_map.ncomp
    so_map: ``so_map``
      the map template that will contain the results of the harmonic transform
    """
    if so_map.ncomp == 1:
        spin = 0
    else:
        spin = [0,2]
    if so_map.pixel == "HEALPIX":
        so_map.data = curvedsky.alm2map_healpix(alms, so_map.data, spin = spin)
    elif so_map.pixel == "CAR":
        so_map.data = curvedsky.alm2map(alms, so_map.data, spin = spin, tweak=tweak)
    else:
        raise ValueError("Map is neither a CAR nor a HEALPIX")
    return so_map

def get_alms(so_map, window, niter, lmax, theta_range=None, dtype=np.complex128):
    """Get a map, multiply by a window and return alms
    This is basically map2alm but with application of the
    window functions.

    Parameters
    ----------

    so_map: ``so_map``
      the data we wants alms from
    window: so_map or tuple of so_map
      a so map with the window function, if the so map has 3 components
      (for spin0 and 2 fields) expect a tuple (window,window_pol)
    theta range: list of 2 elements
      for healpix pixellisation you can specify
      a range [theta_min,theta_max] in radian. All pixel outside this range
      will be assumed to be zero.
    """
    windowed_map = so_map.copy()
    if so_map.ncomp == 3:
        windowed_map.data[0] = so_map.data[0]*window[0].data
        windowed_map.data[1] = so_map.data[1]*window[1].data
        windowed_map.data[2] = so_map.data[2]*window[1].data
    if so_map.ncomp == 1:
        windowed_map.data = so_map.data * window.data
    alms = map2alm(windowed_map, niter, lmax, theta_range=theta_range, dtype=dtype)
    return alms


def get_pure_alms(so_map, window, niter, lmax):

    """Compute pure alms from maps and window function

    Parameters
    ----------

    so_map: ``so_map``
      the data we wants alms from
    window: so_map or tuple of so_map
      a so map with the window function, if the so map has 3 components
      (for spin0 and 2 fields) expect a tuple (window,window_pol)
    niter: integer
      the number of iteration performed while computing the alm
      not that for CAR niter=0 should be enough
    lmax:  integer
      the maximum multipole of the transform

    """

    w1_plus, w1_minus, w2_plus, w2_minus = so_window.get_spinned_windows(window[1], lmax, niter=niter)
    p2 = np.array([window[1].data * so_map.data[1], window[1].data * so_map.data[2]])
    p1 = np.array([(w1_plus.data * so_map.data[1] + w1_minus.data * so_map.data[2]), (w1_plus.data * so_map.data[2] - w1_minus.data * so_map.data[1])])
    p0 = np.array([(w2_plus.data * so_map.data[1] + w2_minus.data * so_map.data[2]), (w2_plus.data * so_map.data[2] - w2_minus.data * so_map.data[1])])

    if so_map.pixel == "CAR":
        p0 = enmap.samewcs(p0,so_map.data)
        p1 = enmap.samewcs(p1,so_map.data)
        p2 = enmap.samewcs(p2,so_map.data)

        alm = curvedsky.map2alm(so_map.data[0] * window[0].data, lmax=lmax)
        s2eblm = curvedsky.map2alm(p2, spin=2, lmax=lmax)
        s1eblm = curvedsky.map2alm(p1, spin=1, lmax=lmax)
        s0eblm = s1eblm.copy()
        s0eblm[0] = curvedsky.map2alm(p0[0], spin=0, lmax=lmax)
        s0eblm[1] = curvedsky.map2alm(p0[1], spin=0, lmax=lmax)

    if so_map.pixel == "HEALPIX":
        alm = hp.sphtfunc.map2alm(so_map.data[0] * window[0].data, lmax=lmax, iter=niter)#curvedsky.map2alm_healpix(map.data[0]*window[0].data,lmax= lmax)
        s2eblm = curvedsky.map2alm_healpix(p2, spin=2, lmax=lmax)
        s1eblm = curvedsky.map2alm_healpix(p1, spin=1, lmax=lmax)
        if niter != 0:
            p2_copy = p2.copy()
            p1_copy = p1.copy()
            for _ in range(niter):
                s2eblm += curvedsky.map2alm_healpix(p2-curvedsky.alm2map_healpix(s2eblm, p2_copy, spin=2), lmax=lmax, spin=2)
                s1eblm += curvedsky.map2alm_healpix(p1-curvedsky.alm2map_healpix(s1eblm, p1_copy, spin=1), lmax=lmax, spin=1)

        s0eblm= s1eblm.copy()
        s0eblm[0] = hp.sphtfunc.map2alm(p0[0], lmax=lmax, iter=niter)
        s0eblm[1] = hp.sphtfunc.map2alm(p0[1], lmax=lmax, iter=niter)

    ell = np.arange(lmax+1)
    filter_1 = np.zeros(lmax+1)
    filter_2 = np.zeros(lmax+1)

    filter_1[2:] = 2 * np.sqrt(1.0 / ((ell[2:] + 2.) * (ell[2:] - 1.)))
    filter_2[2:] = np.sqrt(1.0  / ((ell[2:] + 2.) * (ell[2:] + 1.) * ell[2:] * (ell[2:] - 1.)))

    for k in range(2):
        s1eblm[k] = hp.almxfl(s1eblm[k],filter_1)
        s0eblm[k] = hp.almxfl(s0eblm[k],filter_2)

    elm_p = s2eblm[0] + s1eblm[0] + s0eblm[0]
    blm_b = s2eblm[1] + s1eblm[1] + s0eblm[1]

    return np.array([alm,elm_p,blm_b])



"""
Routines for power spectra estimation and debiasing.
"""
import healpy as hp, numpy as np
from pspy import pspy_utils,so_mcm
from pixell import curvedsky

def get_spectra(alm1, alm2=None, spectra=None):
    """Get the power spectrum of alm1 and alm2, we use healpy.alm2cl for doing this.
    for the spin0 and spin2 case it is a bit ugly as we have to deal with healpix convention.
    Our  convention for spectra is:  ['TT','TE','TB','ET','BT','EE','EB','BE','BB']
    while healpix convention is to take alm1,alm2 and return ['TT','EE','BB','TE','EB','TB']
    
    Parameters
    ----------
    alm1: 1d array
      the spherical harmonic transform of map1
    alm2: 1d array
      the spherical harmonic transform of map2
    spectra: list of strings
      needed for spin0 and spin2 cross correlation, the arrangement of the spectra
      
    Return
    ----------

    The function returns the multipole array l and cl a 1d power spectrum array or
    cl_dict (for spin0 and spin2) a dictionnary of cl with entry spectra
    """

    if spectra is None:
        if alm2 is None:
            cls = hp.sphtfunc.alm2cl(alm1)
        else:
            cls = hp.sphtfunc.alm2cl(alm1, alm2)
        l = np.arange(len(cls))
        return l, cls
        
    cls = hp.sphtfunc.alm2cl(alm1, alm2)
    l = np.arange(len(cls[0]))
    """ spectra_healpix=[TT,EE,BB,TE,EB,TB] """
    spectra_healpix = [spectra[0], spectra[5], spectra[8], spectra[1], spectra[6], spectra[2]]
    cl_dict = {spec: cls[i] for i, spec in enumerate(spectra_healpix)}

    if alm2 is None:
        #here we set ET=TE, BE=EB and BT=TB
        cl_dict[spectra[3]] = cl_dict[spectra[1]]
        cl_dict[spectra[7]] = cl_dict[spectra[6]]
        cl_dict[spectra[4]] = cl_dict[spectra[2]]
    else:
        #here we need to recompute cls inverting the order of the alm to get ET,BT and BE
        cls = hp.sphtfunc.alm2cl(alm2, alm1)
        # spectra_healpix=[TT,EE,BB,ET,BE,BT]
        spectra_healpix = [spectra[0], spectra[5], spectra[8], spectra[3], spectra[7], spectra[4]]
        for i, spec in enumerate(spectra_healpix):
            cl_dict[spec] = cls[i]

    return l, cl_dict
    
    
def get_spectra_pixell(alm1, alm2=None, spectra=None):
    """Get the power spectrum of alm1 and alm2, we use pixell.alm2cl (this is faster)
    Parameters
    ----------
    alm1: 1d array
      the spherical harmonic transform of map1
    alm2: 1d array
      the spherical harmonic transform of map2
    spectra: list of strings
    needed for spin0 and spin2 cross correlation, the arrangement of the spectra
  
    Return
    ----------
    
    The function returns the multipole array l and cl a 1d power spectrum array or
    cl_dict (for spin0 and spin2) a dictionnary of cl with entry spectra
    """
    
    if spectra is None:
        if alm2 is None:
            cls = curvedsky.alm2cl(alm1)
        else:
            cls = curvedsky.alm2cl(alm1, alm2)
        l = np.arange(len(cls))
        return l, cls
        
        
    cls = curvedsky.alm2cl(alm1[:,None], alm2[None,:])
    l = np.arange(len(cls[0,0]))
    cl_dict = {}
    for i, l1 in enumerate(["T","E","B"]):
        for j, l2 in enumerate(["T","E","B"]):
            cl_dict[l1+l2] = cls[i,j]
            
    return(l, cl_dict)


def deconvolve_mode_coupling_matrix(l, cl, inv_mode_coupling_matrix, spectra=None):
    """deconvolve the mode coupling matrix
    Parameters
    ----------
    l: 1d array
      the multipoles or the location of the center of the bins
    cl: 1d array or dict of 1d array
      the power spectra, can be a 1d array (spin0) or a dictionnary (spin0 and spin2)
    inv_mode_coupling_matrix: 2d array
      the inverse of the  mode coupling matrix can be binned or not
    spectra: list of string
        needed for spin0 and spin2 cross correlation, the arrangement of the spectra

    """
    
    n_element = len(l)
    
    if spectra is None:
        ps = np.dot(inv_mode_coupling_matrix, cl)
    else:
        inv_mode_coupling_matrix = so_mcm.coupling_dict_to_array(inv_mode_coupling_matrix)
        vec = []
        for f in spectra:
            vec = np.append(vec, cl[f])
        vec = np.dot(inv_mode_coupling_matrix, vec)
        ps = vec2spec_dict(n_element, vec, spectra)
    
    return l, ps


def bin_spectra(l, cl, binning_file, lmax, type, spectra=None, mbb_inv=None, binned_mcm=True):

    """Bin the power spectra according to a binning file and optionnaly deconvolve the (binned) mode coupling matrix
    
    Parameters
    ----------
    l: 1d array
      the multipoles
    cl: 1d array or dict of 1d array
      the power spectra to bin, can be a 1d array (spin0) or a dictionnary (spin0 and spin2)
    binning_file: data file
      a binning file with format bin low, bin high, bin mean
    lmax: int
      the maximum multipole to consider
    type: string
      the type of binning, either bin Cl or bin Dl
    spectra: list of string
      needed for spin0 and spin2 cross correlation, the arrangement of the spectra
    mbb_inv: 2d array
      optionnaly apply the inverse of the  mode coupling matrix to debiais the spectra
    binned_mcm: boolean
      if mbb_inv is not None, specify if it's binned or not
      
    Return
    ----------
      
    The function return the binned multipole array bin_c and a 1d power spectrum
    array (or dictionnary of 1d power spectra if spectra is not None).
    """

    bin_lo, bin_hi, lb, bin_size = pspy_utils.read_binning_file(binning_file, lmax)
    n_bins = len(bin_hi)

    # the alm2cl return cl starting at l = 0, we use spectra from l = 2
    # this is due in particular to the fact that the mcm is computed only for l>=2
    
    l = np.arange(2, lmax)
    if spectra is None: cl = cl[l]
    else: cl = {f: cl[f][l] for f in spectra}
    
    if type == "Dl": fac = (l * (l + 1) / (2 * np.pi))
    elif type == "Cl": fac = l * 0 + 1

    # we have the option to deconvolve the l-by-l mode coupling matrix
    if (mbb_inv is not None) & (binned_mcm == False):
        l, cl = deconvolve_mode_coupling_matrix(l, cl, mbb_inv, spectra)
    
    # Now the binning part
    if spectra is None:
        ps = np.zeros(n_bins)
        for ibin in range(n_bins):
            loc = np.where((l >= bin_lo[ibin]) & (l <= bin_hi[ibin]))
            ps[ibin] = (cl[loc] * fac[loc]).mean()
    else:
        vec = []
        for f in spectra:
            binned_power = np.zeros(n_bins)
            for ibin in range(n_bins):
                loc = np.where((l >= bin_lo[ibin]) & (l <= bin_hi[ibin]))
                binned_power[ibin] = (cl[f][loc] * fac[loc]).mean()
            vec = np.append(vec, binned_power)
        ps = vec2spec_dict(n_bins, vec, spectra)

    # we have the option to deconvolve the binned mode coupling matrix
    if (mbb_inv is not None) & (binned_mcm == True):
        lb, ps = deconvolve_mode_coupling_matrix(lb, ps, mbb_inv, spectra)
    return lb, ps

def vec2spec_dict(n_bins, vec, spectra):
    """Take a vector of power spectra and return a power spectra dictionnary.
    vec should be of the form [spectra[0], spectra[1], ... ].
    For example [cl_TT,cl_TE,cl_ET, ...., cl_BB]
    
    Parameters
    ----------
    n_bins: int
      the number of bins per spectrum
    vec: 1d array
      an array containing a vector of spectra
    spectra: list of strings
      the arrangement of the spectra for example:
      ['TT','TE','TB','ET','BT','EE','EB','BE','BB']
    """
    return {spec: vec[i * n_bins:(i + 1) * n_bins] for i, spec in enumerate(spectra)}


def write_ps(file_name, l, ps, type, spectra=None):
    """Write down the power spectra to disk.
    
    Parameters
    ----------
    file_name: str
      the name of the file to write the spectra
    l: 1d array
      the multipoles (or binned multipoles)
    ps: 1d array or dict of 1d array
      the power spectrum, if spectra is not None, expect a dictionary with entry spectra
    type:  string
      'Cl' or 'Dl'
    spectra: list of strings
      needed for spin0 and spin2 cross correlation, the arrangement of the spectra
    """

    if spectra is None:
        ps_list = [ps]
        ps_list[0:0] = [l]
        str = "l"
        str += " %s"%type
        ps_list = np.array(ps_list)
        np.savetxt(file_name, np.transpose(ps_list), header=str)
    else:
        ps_list = [ps[f] for f in spectra]
        ps_list[0:0] = [l]
        str = "l"
        for l1 in spectra:
            str += " %s_%s"%(type, l1)
        ps_list = np.array(ps_list)
        np.savetxt(file_name, np.transpose(ps_list), header=str)


def read_ps(file_name, spectra=None):
    """Read the power spectra.
        
    Parameters
    ----------
    file_name: str
      the name of the file to read the spectra
    spectra: list of strings
      needed for spin0 and spin2 cross correlation, the arrangement of the spectra
      
    Return
    ----------
      
    The function return the multipole l (or binned multipoles) and a 1d power spectrum
    array or a dictionnary of power spectra (if spectra is not None).
    """
    
    data = np.loadtxt(file_name)
    if spectra is None:
        return data[:, 0], data[:, 1]

    l = data[:, 0]
    ps = {spec: data[:, i + 1] for i, spec in enumerate(spectra)}
    return l, ps

def write_ps_hdf5(file, spec_name, l, ps, spectra=None):
    """Write down the power spectra in a hdf5 file.
        
    Parameters
    ----------
    file: hdf5
      the name of the hdf5 file
    spec_name: string
      the name of the group in the hdf5 file
    l: 1d array
      the multipoles (or binned multipoles)
    ps: 1d array or dict of 1d array
      the power spectrum, if spectra is not None, expect a dictionary with entry spectra
    spectra: list of strings
      needed for spin0 and spin2 cross correlation, the arrangement of the spectra
    """

    def array_from_dict(l,ps,spectra=None):
        array = []
        array += [l]
        if spectra == None:
            array += [ps]
        else:
            for spec in spectra:
                array += [ps[spec]]
        return array
    
    group = file.create_group(spec_name)
    array = array_from_dict(l, ps, spectra=spectra)
    group.create_dataset(name="data", data=array, dtype="float")

def read_ps_hdf5(file, spec_name, spectra=None):
    """Read the power spectra in a hdf5 file.
        
    Parameters
    ----------
    file: hdf5
      the name of the hdf5 file
    spec_name: string
      the name of the group in the hdf5 file
    spectra: list of strings
      needed for spin0 and spin2 cross correlation, the arrangement of the spectra

    Return
    ----------

    The function returns the multipole l (or binned multipoles) and a 1d power spectrum
    array (or a dictionnary of power spectra if spectra is not None).

    """
    
    spec = file[spec_name]
    data = np.array(spec["data"]).T
    
    l = data[:,0]
    if spectra is None:
        ps = data[:, 1]
    else:
        ps = {spec: data[:, i + 1] for i, spec in enumerate(spectra)}

    return l, ps





"""
Routines for mode coupling calculation. For more details on computation of the matrix see
https://pspy.readthedocs.io/en/latest/scientific_doc.pdf.
"""

from copy import deepcopy

import healpy as hp
import numpy as np

from pspy import pspy_utils, so_cov, sph_tools
from pspy.mcm_fortran.mcm_fortran import mcm_compute as mcm_fortran

def mcm_and_bbl_spin0(win1,
                      binning_file,
                      lmax,
                      niter,
                      type="Dl",
                      win2=None,
                      bl1=None,
                      bl2=None,
                      input_alm=False,
                      binned_mcm=True,
                      save_file=None,
                      l_exact=None,
                      l_toep=None,
                      l_band=None,
                      l3_pad=2000,
                      return_coupling_only=False):
                      
                      
    """Get the mode coupling matrix and the binning matrix for spin0 fields
     Parameters
     ----------
     win1: so_map (or alm)
       the window function of survey 1, if input_alm=True, expect wlm1
     binning_file: text file
       a binning file with three columns bin low, bin high, bin mean
     lmax: integer
       the maximum multipole to consider for the spectra computation
     type: string
       the type of binning, either bin Cl or bin Dl
     win2: so_map (or alm)
       the window function of survey 2, if input_alm=True, expect wlm2
     bl1: 1d array
       the beam of survey 1, expected to start at l=0
     bl2: 1d array
       the beam of survey 2, expected to start at l=0
     niter: int
       specify the number of iteration in map2alm
     binned_mcm: boolean
       wether to bin or not the mcm
     save_file: boolean
       save the mcm and bbl to disk
     l_toep: int
     l_band: int
     l_exact: int

     """

    if type == "Dl": doDl = 1
    if type == "Cl": doDl = 0
    if type not in ["Dl", "Cl"]:
        raise ValueError("Unkown 'type' value! Must be either 'Dl' or 'Cl'")

    if input_alm == False:
        l_max_limit = win1.get_lmax_limit()
        if lmax > l_max_limit: raise ValueError("the requested lmax is too high with respect to the map pixellisation")
        maxl = np.minimum(lmax + l3_pad, l_max_limit)

        win1 = sph_tools.map2alm(win1, niter=niter, lmax=maxl)
        if win2 is not None:
            win2 = sph_tools.map2alm(win2, niter=niter, lmax=maxl)

    if win2 is None:
        wcl = hp.alm2cl(win1)
    else:
        wcl = hp.alm2cl(win1, win2)

    l = np.arange(len(wcl))
    wcl *= (2 * l + 1)

    if bl1 is None: bl1 = np.ones(len(l)+2)
    if bl2 is None: bl2 = bl1.copy()

    mcm = np.zeros((lmax, lmax))

    if l_toep is None: l_toep = lmax
    if l_band is None: l_band = lmax
    if l_exact is None: l_exact = lmax

    mcm_fortran.calc_coupling_spin0(wcl,
                                    l_exact,
                                    l_band,
                                    l_toep,
                                    mcm.T)

    if l_toep < lmax:
        mcm = format_toepliz_fortran2(mcm, l_toep, l_exact, lmax)

    mcm_fortran.fill_upper(mcm.T)

    if return_coupling_only == True:
        return mcm[:lmax - 2, :lmax - 2]

    fac = (2 * np.arange(2, lmax + 2) + 1) / (4 * np.pi) * bl1[2:lmax + 2] * bl2[2:lmax + 2]
    mcm *= fac

    bin_lo, bin_hi, bin_c, bin_size = pspy_utils.read_binning_file(binning_file, lmax)
    n_bins = len(bin_hi)
    
    Bbl = np.zeros((n_bins, lmax))

    if binned_mcm == True:
        mbb = np.zeros((n_bins, n_bins))
        mcm_fortran.bin_mcm(mcm.T, bin_lo, bin_hi, bin_size, mbb.T, doDl)
        mcm_fortran.binning_matrix(mcm.T, bin_lo, bin_hi, bin_size, Bbl.T, doDl)
        mbb_inv = np.linalg.inv(mbb)
        Bbl = np.dot(mbb_inv, Bbl)
        if save_file is not None:
            save_coupling(save_file, mbb_inv, Bbl)
        return mbb_inv, Bbl
    else:
        identity_matrix = np.eye(mcm.shape[0])
        mcm_fortran.binning_matrix(identity_matrix.T, bin_lo, bin_hi, bin_size, Bbl.T, doDl)
        mcm = mcm[:lmax - 2, :lmax - 2]
        mcm_inv = np.linalg.inv(mcm)
            
        if save_file is not None:
            save_coupling(save_file, mcm_inv, Bbl)
        return mcm_inv, Bbl


def mcm_and_bbl_spin0and2(win1,
                          binning_file,
                          lmax,
                          niter,
                          type="Dl",
                          win2=None,
                          bl1=None,
                          bl2=None,
                          input_alm=False,
                          pure=False,
                          binned_mcm=True,
                          save_file=None,
                          l3_pad=2000,
                          l_exact=None,
                          l_toep=None,
                          l_band=None,
                          return_coupling_only=False):
                          
    """Get the mode coupling matrix and the binning matrix for spin 0 and 2 fields

    Parameters
    ----------

    win1: python tuple of so_map or alms (if input_alm=True)
      a python tuple (win_spin0,win_spin2) with the window functions of survey 1, if input_alm=True, expect (wlm_spin0, wlm_spin2)
    binning_file: text file
      a binning file with three columns bin low, bin high, bin mean
    lmax: integer
      the maximum multipole to consider
    type: string
      the type of binning, either bin Cl or bin Dl
    win2: python tuple of so_map or alms (if input_alm=True)
      a python tuple (win_spin0,win_spin2) with the window functions of survey 1, if input_alm=True, expect (wlm_spin0, wlm_spin2)
    bl1: python tuple of 1d array
      a python tuple (beam_spin0,beam_spin2) with the beam of survey 1, expected to start at l=0
    bl2: python tuple of 1d array
      a python tuple (beam_spin0,beam_spin2) with the beam of survey 2, expected to start at l=0
    niter: int
      specify the number of iteration in map2alm
    pureB: boolean
      do B mode purification
    binned_mcm: boolean
      wether to bin or not the mcm
    save_file: boolean
      save the mcm and bbl to disk
    l_toep: int
    l_band: int
    l_exact: int

    save_coupling: str
    """
        
    def get_coupling_dict(array):
        ncomp, dim1, dim2 = array.shape
        dict = {}
        dict["spin0xspin0"] = array[0, :, :]
        dict["spin0xspin2"] = array[1, :, :]
        dict["spin2xspin0"] = array[2, :, :]
        dict["spin2xspin2"] = np.zeros((4 * dim1, 4 * dim2))
        for i in range(4):
            dict["spin2xspin2"][i * dim1:(i + 1) * dim1, i * dim2:(i + 1) * dim2] = array[3, :, :]
        dict["spin2xspin2"][2 * dim1:3 * dim1, dim2:2 * dim2] = -array[4, :, :]
        dict["spin2xspin2"][dim1:2 * dim1, 2 * dim2:3 * dim2] = -array[4, :, :]
        dict["spin2xspin2"][3 * dim1:4 * dim1, :dim2] = array[4, :, :]
        dict["spin2xspin2"][:dim1, 3 * dim2:4 * dim2] = array[4, :, :]
        return dict

    if type == "Dl": doDl = 1
    if type == "Cl": doDl = 0
    if type not in ["Dl", "Cl"]:
        raise ValueError("Unkown 'type' value! Must be either 'Dl' or 'Cl'")

    if input_alm == False:
        l_max_limit = win1[0].get_lmax_limit()
        if lmax > l_max_limit: raise ValueError("the requested lmax is too high with respect to the map pixellisation")
        maxl = np.minimum(lmax + l3_pad, l_max_limit)

        win1_alm_T = sph_tools.map2alm(win1[0], niter=niter, lmax=maxl)
        win1_alm_P = sph_tools.map2alm(win1[1], niter=niter, lmax=maxl)
        win1 = (win1_alm_T, win1_alm_P)
        if win2 is not None:
            win2_alm_T = sph_tools.map2alm(win2[0], niter=niter, lmax=maxl)
            win2_alm_P = sph_tools.map2alm(win2[1], niter=niter, lmax=maxl)
            win2 = (win2_alm_T, win2_alm_P)
                                                  
    if win2 is None: win2 = deepcopy(win1)
    if bl1 is None: bl1 = (np.ones(2 + lmax), np.ones(2 + lmax))
    if bl2 is None: bl2 = deepcopy(bl1)

    wcl, wbl = {}, {}
    spins = ["spin0", "spin2"]
    spin_pairs = []

    for i, spin1 in enumerate(spins):
        for j, spin2 in enumerate(spins):
            wcl["%sx%s" % (spin1, spin2)] = hp.alm2cl(win1[i], win2[j])
            wcl_lmax = len(wcl["%sx%s" % (spin1, spin2)])
            wcl["%sx%s" % (spin1, spin2)] *= (2 * np.arange(wcl_lmax) + 1)
            wbl["%sx%s" % (spin1, spin2)] = bl1[i][2:lmax + 2] * bl2[j][2:lmax + 2]
            spin_pairs += ["%sx%s" % (spin1, spin2)]
            
    mcm = np.zeros((5, lmax, lmax))

    if pure == False:
        if l_toep is None: l_toep = lmax
        if l_band is None: l_band = lmax
        if l_exact is None: l_exact = lmax

        mcm_fortran.calc_coupling_spin0and2(wcl["spin0xspin0"],
                                            wcl["spin0xspin2"],
                                            wcl["spin2xspin0"],
                                            wcl["spin2xspin2"],
                                            l_exact,
                                            l_band,
                                            l_toep,
                                            mcm.T)

        for id_mcm in range(5):
            if l_toep < lmax:
                mcm[id_mcm] = format_toepliz_fortran2(mcm[id_mcm], l_toep, l_exact, lmax)
            mcm_fortran.fill_upper(mcm[id_mcm].T)
    else:
        mcm_fortran.calc_mcm_spin0and2_pure(wcl["spin0xspin0"],
                                            wcl["spin0xspin2"],
                                            wcl["spin2xspin0"],
                                            wcl["spin2xspin2"],
                                            mcm.T)

    if return_coupling_only == True:
        return mcm[:, :lmax - 2, :lmax - 2]

    for id_mcm, spairs in enumerate(["spin0xspin0", "spin0xspin2", "spin2xspin0", "spin2xspin2", "spin2xspin2"]):
        fac = (2 * np.arange(2, lmax + 2) + 1) / (4 * np.pi) *  wbl[spairs]
        mcm[id_mcm] *= fac

    bin_lo, bin_hi, bin_c, bin_size = pspy_utils.read_binning_file(binning_file, lmax)
    n_bins = len(bin_hi)

    if binned_mcm == True:
        mbb_array = np.zeros((5, n_bins, n_bins))
        Bbl_array = np.zeros((5, n_bins, lmax))
        
        for id_mcm in range(5):

            mcm_fortran.bin_mcm((mcm[id_mcm, :, :]).T,
                                 bin_lo,
                                 bin_hi,
                                 bin_size,
                                 (mbb_array[id_mcm, :, :]).T,
                                 doDl)

            mcm_fortran.binning_matrix((mcm[id_mcm, :, :]).T,
                                        bin_lo,
                                        bin_hi,
                                        bin_size,
                                        (Bbl_array[id_mcm, :, :]).T,
                                        doDl)

        
        mbb = get_coupling_dict(mbb_array)
        Bbl = get_coupling_dict(Bbl_array)

        mbb_inv = {}
        for s in spin_pairs:
            mbb_inv[s] = np.linalg.inv(mbb[s])
            Bbl[s] = np.dot(mbb_inv[s], Bbl[s])
            
        if save_file is not None:
            save_coupling(save_file, mbb_inv, Bbl, spin_pairs=spin_pairs)
        return mbb_inv, Bbl

    else:
        Bbl_array = np.zeros((n_bins, lmax))
        identity_matrix = np.eye(mcm[0].shape[0])
        mcm_fortran.binning_matrix(identity_matrix.T, bin_lo, bin_hi, bin_size, Bbl_array.T, doDl)

        mcm = get_coupling_dict(mcm[:, :lmax - 2, :lmax - 2])
        mcm_inv = {}
        for s in spin_pairs:
            mcm_inv[s] = np.linalg.inv(mcm[s])
            

        # if we deconvolve the full mcm, the binning matrix is the same for all spectra
        Bbl = {}
        Bbl["spin0xspin0"] = Bbl_array
        Bbl["spin0xspin2"] = Bbl_array
        Bbl["spin2xspin0"] = Bbl_array
        Bbl["spin2xspin2"] = np.zeros((4 * n_bins, 4 * lmax))
        for i in range(4):
            Bbl["spin2xspin2"][i * n_bins:(i + 1) * n_bins, i * lmax:(i + 1) * lmax] = Bbl_array
        if save_file is not None:
            save_coupling(save_file, mcm_inv, Bbl, spin_pairs=spin_pairs)
            
        return mcm_inv, Bbl
    

def format_toepliz_fortran(coupling, l_toep, lmax):
    """take a matrix and apply the toepliz appoximation (fortran)
    Parameters
    ----------

    coupling: array
      consist of an array where the upper part is the exact matrix and
      the lower part is the diagonal. We will feed the off diagonal
      of the lower part using the measurement of the correlation from the exact computatio
    l_toep: integer
      the l at which we start the approx
    lmax: integer
      the maximum multipole of the array
    """
    toepliz_array = np.zeros(coupling.shape)
    mcm_fortran.toepliz_array_fortran(toepliz_array.T, coupling.T, l_toep)
    toepliz_array[coupling != 0] = coupling[coupling != 0]
    return toepliz_array

def format_toepliz_fortran2(coupling, l_toep, l_exact, lmax):
    """take a matrix and apply the toepliz appoximation (fortran)
    Parameters
    ----------

    coupling: array
      consist of an array where the upper part is the exact matrix and
      the lower part is the diagonal. We will feed the off diagonal
      of the lower part using the measurement of the correlation from the exact computatio
    l_toep: integer
      the l at which we start the approx
    l_exact: integer
      the l until which we do the exact computation
    lmax: integer
      the maximum multipole of the array
    """
    toepliz_array = np.zeros(coupling.shape)
    mcm_fortran.toepliz_array_fortran2(toepliz_array.T, coupling.T, l_toep, l_exact)
    toepliz_array[coupling != 0] = coupling[coupling != 0]
    return toepliz_array


def coupling_dict_to_array(dict):
    """Take a mcm or Bbl dictionnary with entries:

    -  (spin0xspin0), (spin0xspin2), (spin2xspin0), (spin2xspin2)

    and return a 9 x dim1, 9 x dim2 array.
    dim1 and dim2 are the dimensions of the spin0 object
    """

    dim1, dim2 = dict["spin0xspin0"].shape
    array = np.zeros((9 * dim1, 9 * dim2))
    array[0:dim1, 0:dim2] = dict["spin0xspin0"]
    array[dim1:2 * dim1, dim2:2 * dim2] = dict["spin0xspin2"]
    array[2 * dim1:3 * dim1, 2 * dim2:3 * dim2] = dict["spin0xspin2"]
    array[3 * dim1:4 * dim1, 3 * dim2:4 * dim2] = dict["spin2xspin0"]
    array[4 * dim1:5 * dim1, 4 * dim2:5 * dim2] = dict["spin2xspin0"]
    array[5 * dim1:9 * dim1, 5 * dim2:9 * dim2] = dict["spin2xspin2"]
    
    return array

def apply_Bbl(Bbl, ps, spectra=None):
    """Bin theoretical power spectra

    Parameters
    ----------

    Bbl: 2d array (or dict of 2d array)
        a binning matrix, if spectra is not None will be a Bbl dictionnary for spin0 and 2 fields,
        otherwise a (n_bins,lmax) matrix
    ps: 1d array or dict of 1d array
      theory ps, if spectra is not None it will be a ps dictionnary, otherwise a (lmax) vector
    spectra: list of string
      needed for spin0 and spin2 cross correlation, the arrangement of the spectra
    """

    if spectra is not None:
        Bbl_array = coupling_dict_to_array(Bbl)
        ps_vec = ps[spectra[0]]
        for spec in spectra[1:]:
            ps_vec = np.append(ps_vec, ps[spec])
        ps_b = np.dot(Bbl_array, ps_vec)
        n_bins = int(Bbl_array.shape[0] / 9)
        ps_th = {}
        for i, spec in enumerate(spectra):
            ps_th[spec] = ps_b[i * n_bins:(i + 1) * n_bins]
    else:
        ps_th = np.dot(Bbl, ps)
    return ps_th

def save_coupling(prefix, mode_coupling_inv, Bbl, spin_pairs=None):
    """Save the inverse of the mode coupling matrix and the binning matrix in npy format

    Parameters
    ----------

    prefix: string
      the prefix for the name of the file
    mode_coupling_inv: 2d array (or dict of 2d array)
      the inverse of the mode coupling matrix, if spin pairs is not none, should be a dictionnary with entries
      -  spin0xspin0, spin0xspin2, spin2xspin0, spin2xspin2
    Bbl: 2d array (or dict of 2d array)
      the binning matrix,if spin pairs is not none, should be a dictionnary with entries
      -  spin0xspin0, spin0xspin2, spin2xspin0, spin2xspin2
      otherwise, it will be a single matrix.
    spin_pairs: list of strings
      needed for spin0 and 2 fields.
    """

    if spin_pairs is not None:
        for spin in spin_pairs:
            np.save(prefix + "_mode_coupling_inv_%s.npy" % spin, mode_coupling_inv[spin])
            np.save(prefix + "_Bbl_%s.npy" % spin, Bbl[spin])
    else:
        np.save(prefix + "_mode_coupling_inv.npy", mode_coupling_inv)
        np.save(prefix + "_Bbl.npy", Bbl)



def read_coupling(prefix, spin_pairs=None):
    """Read the inverse of the mode coupling matrix and the binning matrix
    Parameters
    ----------

    prefix: string
      the prefix for the name of the file
    spin_pairs: list of strings
      needed for spin0 and 2 fields.
    unbin: boolean
      also read the unbin matrix
    """

    if spin_pairs is not None:
        Bbl = {}
        mode_coupling_inv = {}
        for spin in spin_pairs:
            mode_coupling_inv[spin] = np.load(prefix + "_mode_coupling_inv_%s.npy" % spin)
            Bbl[spin] = np.load(prefix + "_Bbl_%s.npy" % spin)
    else:
        mode_coupling_inv = np.load(prefix + "_mode_coupling_inv.npy")
        Bbl = np.load(prefix + "_Bbl.npy")
    
    return mode_coupling_inv, Bbl



"""
SO map class for handling healpix and CAR maps.
This is a wrapper around healpix and enlib (pixell).
"""

import os
from copy import deepcopy

import astropy.io.fits as pyfits
import healpy as hp
import matplotlib.pyplot as plt
import numpy as np
from pixell import colorize, curvedsky, enmap, enplot, powspec, reproject

from pspy.pspy_utils import ps_lensed_theory_to_dict
from pspy.sph_tools import map2alm, alm2map


class so_map:
    """Class defining a ``so_map`` object."""

    def __init__(self):
        self.pixel = None
        self.nside = None
        self.ncomp = None
        self.data = None
        self.geometry = None
        self.coordinate = None

    def copy(self):
        """Create a copy of the ``so_map`` object."""

        return deepcopy(self)

    def info(self):
        """Print information about the ``so_map`` object."""

        print("pixellisation:", self.pixel)
        print("number of components:", self.ncomp)
        print("number of pixels:", self.data.shape[:] if self.ncomp == 1 else self.data.shape[1:])
        print("nside:", self.nside)
        print("geometry:", self.geometry)
        print("coordinates:", self.coordinate)

    def write_map(self, file_name):
        """Write the ``so_map`` to disk.

        Parameters
        ----------
        filename : string
          the name of the fits file
        """

        if self.pixel == "HEALPIX":
            hp.fitsfunc.write_map(file_name, self.data, overwrite=True)
        if self.pixel == "CAR":
            enmap.write_map(file_name, self.data)

    def calibrate(self, cal=1.0, pol_eff=1.0):
        """Calibrate the ``so_map``, also optionnaly apply a pol efficiency
        Parameters
        ----------
        cal : float
            the calibration factor to apply
        pol_eff : float
            the polarisation efficiency
        """
        
        if self.ncomp == 1:
            self.data *= cal
        else:
            self.data[0] *= cal
            self.data[1] *= cal * pol_eff
            self.data[2] *= cal * pol_eff

        return self
    
    def upgrade(self, factor):
        """Upgrade the ``so_map``.

        Parameters
        ----------
        factor : integer
          factor of increased pixel resolution (should be a factor of 2)

        """

        assert factor % 2 == 0, "factor should be a factor of 2"

        upgrade = self.copy()
        if self.pixel == "HEALPIX":
            nside_out = int(self.nside * factor)
            upgrade.data = hp.pixelfunc.ud_grade(self.data, nside_out=nside_out)
            upgrade.nside = nside_out
        if self.pixel == "CAR":
            upgrade.data = enmap.upgrade(self.data, factor)
            upgrade.geometry = upgrade.data.geometry[1:]
        return upgrade

    def downgrade(self, factor):
        """Downgrade the ``so_map``.

        Parameters
        ----------
        factor : integer
          factor of decreased pixel resolution (should be a factor of 2)

        """

        assert factor % 2 == 0, "factor should be a factor of 2"

        downgrade = self.copy()
        if self.pixel == "HEALPIX":
            nside_out = int(self.nside / factor)
            downgrade.data = hp.pixelfunc.ud_grade(self.data, nside_out=nside_out)
            downgrade.nside = nside_out
        if self.pixel == "CAR":
            downgrade.data = enmap.downgrade(self.data, factor)
            downgrade.geometry = downgrade.data.geometry[1:]
        return downgrade

    def synfast(self, clfile):
        """Fill a ``so_map`` with a cmb gaussian simulation.

        Parameters
        ----------
        clfile : CAMB data file
          lensed power spectra file from CAMB

        """

        synfast = self.copy()
        if self.pixel == "HEALPIX":
            l, ps = ps_lensed_theory_to_dict(clfile, output_type="Cl", start_at_zero=True)
            if self.ncomp == 1:
                synfast.data = hp.sphtfunc.synfast(ps["TT"], self.nside, new=True)
            else:
                synfast.data = hp.sphtfunc.synfast(
                    (ps["TT"], ps["EE"], ps["BB"], ps["TE"]), self.nside, new=True
                )

        if self.pixel == "CAR":
            ps = powspec.read_spectrum(clfile)[: self.ncomp, : self.ncomp]
            synfast.data = curvedsky.rand_map(self.data.shape, self.data.wcs, ps)

        return synfast

    def get_lmax_limit(self):
        """Return the maximum lmax corresponding to the ``so_map`` pixellisation"""

        if self.pixel == "HEALPIX":
            l_max_limit = 3 * self.nside - 1
        elif self.pixel == "CAR":
            cdelt = self.data.wcs.wcs.cdelt[1]
            l_max_limit = 360 / cdelt / 4
        return l_max_limit
        
    def get_pixwin(self):
        """compute the pixel window function corresponding to the map pixellisation
        """
        if self.pixel == "HEALPIX":
            pixwin = hp.pixwin(self.nside)
        if self.pixel == "CAR":
            wy, wx = enmap.calc_window(self.data.shape)
            pixwin = (wy[:,None] * wx[None,:])
        return pixwin
            
    def convolve_with_pixwin(self, niter=3, binary=None, pixwin=None):
        """Convolve a ``so_map`` object with a pixel window function
        The convolution is done in harmonics space, for CAR maps
        the pixwin is anisotropic (the pixel varies in size across the maps)
        and the convolution is done in Fourier space.
        We optionaly apply a binary before doing the operation to remove pathological pixels, note
        that this operation is dangerous since we do harmonic transform of a masked map.
        
        Parameters
        ----------
        niter: integer
          the number of iteration performed while computing the alm
          not that for CAR niter=0 should be enough
        binary: so_map
            a binary mask that remove pathological pixel before doing the harmonic
            space operation
        pixwin: 1d array for healpix, 2d array for CAR
            this allow you to pass a precomputed pixel window function
        """

        lmax = self.get_lmax_limit()
            
        if binary is not None: self.data *= binary.data
        if pixwin is None: pixwin = self.get_pixwin()

        if self.pixel == "HEALPIX":
            alms = map2alm(self, niter, lmax)
            alms = curvedsky.almxfl(alms, pixwin)
            self = alm2map(alms, self)
        if self.pixel == "CAR":
            self = fourier_convolution(self, pixwin)
            
        return self
            
    def subtract_mean(self, mask=None):
        """Subtract mean from a ``so_map`` object, optionnaly within a mask

        Parameters
        ----------
        mask: either a single so_map (for ncomp = 1) or a tuple of SO map e.g (mask_T, mask_P)
        """
            
        if mask is None:
            self.data -= np.mean(self.data)
        else:
            if self.ncomp == 1:
                self.data -= np.mean(self.data * mask.data)
            else:
                self.data[0] -= np.mean(self.data * mask[0].data)
                self.data[1] -= np.mean(self.data * mask[1].data)
                self.data[2] -= np.mean(self.data * mask[1].data)
            
        return self
            
    def subtract_mono_dipole(self, mask=None, bunch=24):
        """Subtract monopole and dipole from a ``so_map`` object.

        Parameters
        ----------
        mask: either a single so_map (for ncomp = 1) or a tuple of SO map e.g (mask_T, mask_P)
        bunch: int
            the bunch size (default: 24)
        """

        if self.ncomp == 1:
            self.data = subtract_mono_dipole(
                emap=self.data,
                mask=None if mask is None else mask.data,
                healpix=self.pixel == "HEALPIX",
                bunch=bunch,
            )
        else:
            for i in range(self.ncomp):
                self.data[i] = subtract_mono_dipole(
                    emap=self.data[i],
                    mask=None if mask is None else mask[i if i < 2 else 1].data,
                    healpix=self.pixel == "HEALPIX",
                    bunch=bunch,
                )

    def plot(
        self,
        color="planck",
        color_range=None,
        file_name=None,
        ticks_spacing_car=1,
        title="",
        cbar=True,
        hp_gnomv=None,
    ):
        """Plot a ``so_map``.

        Parameters
        ----------
        color: cmap
          a matplotlib colormap (or 'planck')
        color_range: scalar for single component or len(3) list for T,Q,U.
          the range of the colorscale
        file_name: string
          file_name is the name of the png file that will be created, if None the plot
          will be displayed.
        title: string
          the title of the plot.
        cbar: boolean
          set to True to display the colorbar.
        ticks_spacing_CAR: float
          for CAR plot, choose the spacing of the ticks.
        hp_gnomv: tuple
          gnomview projection for HEALPIX plotting, expected (lon_c,lat_c,xsize,reso).

        """
        try:
            colorize.mpl_setdefault(color)
        except KeyError:
            if self.pixel == "CAR":
                raise KeyError(
                    "Color name must be a pixell color map name {}!".format(
                        list(colorize.schemes.keys())
                    )
                )

        if self.pixel == "HEALPIX":
            cmap = plt.get_cmap(color)
            cmap.set_bad("white")
            cmap.set_under("white")
            if self.ncomp == 1:

                min_range = -color_range if color_range is not None else None
                max_range = +color_range if color_range is not None else None

                if hp_gnomv is not None:
                    lon, lat, xsize, reso = hp_gnomv
                    hp.gnomview(
                        self.data,
                        min=min_range,
                        max=max_range,
                        cmap=cmap,
                        notext=True,
                        title=title,
                        cbar=cbar,
                        rot=(lon, lat, 0),
                        xsize=xsize,
                        reso=reso,
                    )
                else:
                    hp.mollview(
                        self.data,
                        min=min_range,
                        max=max_range,
                        cmap=cmap,
                        notext=True,
                        title=title,
                        cbar=cbar,
                    )
                if file_name is not None:
                    plt.savefig(file_name + ".png", bbox_inches="tight")
                    plt.clf()
                    plt.close()
                else:
                    plt.show()
            else:
                fields = ["T", "Q", "U"]
                min_ranges = {field: None for field in fields}
                max_ranges = {field: None for field in fields}
                if color_range is not None:
                    for i, field in enumerate(fields):
                        min_ranges[field] = -color_range[i]
                        max_ranges[field] = +color_range[i]

                for data, field in zip(self.data, fields):
                    if hp_gnomv is not None:
                        lon, lat, xsize, reso = hp_gnomv
                        hp.gnomview(
                            data,
                            min=min_ranges[field],
                            max=max_ranges[field],
                            cmap=cmap,
                            notext=True,
                            title=field + "" + title,
                            cbar=cbar,
                            rot=(lon, lat, 0),
                            xsize=xsize,
                            reso=reso,
                        )
                    else:
                        hp.mollview(
                            data,
                            min=min_ranges[field],
                            max=max_ranges[field],
                            cmap=cmap,
                            notext=True,
                            title=field + "" + title,
                            cbar=cbar,
                        )
                    if file_name is not None:
                        plt.savefig(file_name + "_%s" % field + ".png", bbox_inches="tight")
                        plt.clf()
                        plt.close()
                    else:
                        plt.show()

        if self.pixel == "CAR":
            if self.ncomp == 1:
                if color_range is not None:
                    max_range = "%s" % (color_range)
                else:
                    max_range = "%s" % (np.max(self.data))

                plots = enplot.get_plots(
                    self.data, color=color, range=max_range, colorbar=1, ticks=ticks_spacing_car
                )

                for plot in plots:
                    if file_name is not None:
                        enplot.write(file_name + ".png", plot)
                    else:
                        plot.img.show()

            if self.ncomp == 3:
                fields = ["T", "Q", "U"]

                if color_range is not None:
                    max_range = "%s:%s:%s" % (color_range[0], color_range[1], color_range[2])
                else:
                    max_range = "%s:%s:%s" % (
                        np.max(self.data[0]),
                        np.max(self.data[1]),
                        np.max(self.data[2]),
                    )

                plots = enplot.get_plots(
                    self.data, color=color, range=max_range, colorbar=1, ticks=ticks_spacing_car
                )

                for (plot, field) in zip(plots, fields):
                    if file_name is not None:
                        enplot.write(file_name + "_%s" % field + ".png", plot)
                    else:
                        # enplot.show(plot,method="ipython")
                        plot.img.show()




def read_map(file, coordinate=None, fields_healpix=None, car_box=None, geometry=None):
    """Create a ``so_map`` object from a fits file.

    Parameters
    ----------
    file: fits file
      name of the fits file
    coordinate: string
      coordinate system of the map
    fields_healpix: integer
      if fields_healpix is not None, load the specified field
    car_box:  2x2 array
        [[dec0,ra0],[dec1,ra1]] in degree

    """

    new_map = so_map()
    hdulist = pyfits.open(file)
    try:
        header = hdulist[1].header
        new_map.pixel = "HEALPIX"
        if fields_healpix is None:
            new_map.ncomp = header["TFIELDS"]
            new_map.data = hp.fitsfunc.read_map(file, field=np.arange(new_map.ncomp))
        else:
            try:
                new_map.ncomp = len(fields_healpix)
            except:
                new_map.ncomp = 1
            new_map.data = hp.fitsfunc.read_map(file, field=fields_healpix)

        new_map.nside = hp.pixelfunc.get_nside(new_map.data)
        new_map.geometry = "healpix geometry"
        try:
            new_map.coordinate = header["SKYCOORD"]
        except:
            new_map.coordinate = None

    except:
        header = hdulist[0].header
        new_map.pixel = header["CTYPE1"][-3:]
        try:
            new_map.ncomp = header["NAXIS3"]
        except:
            new_map.ncomp = 1

        if car_box is not None:
            car_box = np.array(car_box) * np.pi / 180
            new_map.data = enmap.read_map(file, box=car_box)
        elif geometry is not None:
            new_map.data = enmap.read_map(file, geometry=geometry)
        else:
            new_map.data = enmap.read_map(file)

        new_map.nside = None
        new_map.geometry = new_map.data.geometry[1:]
        new_map.coordinate = header["RADESYS"]
        if new_map.coordinate == "ICRS":
            new_map.coordinate = "equ"

    hdulist.close()

    if coordinate is not None:
        new_map.coordinate = coordinate

    return new_map


def from_components(T, Q, U):
    """Create a (T,Q,U) ``so_map`` object from three fits files.

    Parameters
    ----------
    T : fits file
      name of the T fits file
    Q : fits file
      name of the Q fits file
    U : fits file
      name of the U fits file
    """

    ncomp = 3
    T = enmap.read_map(T)
    Q = enmap.read_map(Q)
    U = enmap.read_map(U)
    shape, wcs = T.geometry
    shape = (ncomp,) + shape
    new_map = so_map()
    new_map.data = enmap.zeros(shape, wcs=wcs, dtype=None)
    new_map.data[0] = T
    new_map.data[1] = Q
    new_map.data[2] = U
    new_map.pixel = "CAR"
    new_map.nside = None
    new_map.ncomp = ncomp
    new_map.geometry = T.geometry[1:]
    new_map.coordinate = "equ"

    return new_map
    
def get_submap_car(map_car, box, mode="round"):
    """Cut a CAR submap (using pixell).

    Parameters
    ----------
    map : CAR map in ``so_map`` format
      the map to be cut
    box : array_like
      The [[fromy,fromx],[toy,tox]] bounding box to select.
      The resulting map will have a bounding box as close
      as possible to this, but will differ slightly due to
      the finite pixel size.

    mode : str
      How to handle partially selected pixels:
      "round": round bounds using standard rules
      "floor": both upper and lower bounds will be rounded down
      "ceil":  both upper and lower bounds will be rounded up
      "inclusive": lower bounds are rounded down, and upper bounds up
      "exclusive": lower bounds are rounded up, and upper bounds down"""

    submap = map_car.copy()
    submap.data = map_car.data.submap(box, mode=mode)
    submap.geometry = map_car.data.submap(box, mode=mode).geometry[1:]

    return submap


def get_box(ra0, ra1, dec0, dec1):
    """Create box in equatorial coordinates.

    Parameters
    ----------
    ra0, dec0, ra1, dec1 : floats
      coordinates of the box in degrees
    """

    box = np.array([[dec0, ra1], [dec1, ra0]]) * np.pi / 180

    return box


def bounding_box_from_map(map_car):
    """Get a coordinate box from a map.

    Parameters
    ----------
    map_car : ``so_map`` in CAR coordinates
      the map used to define the box
    """

    shape, wcs = map_car.data.geometry

    return enmap.box(shape, wcs)


def from_enmap(emap):
    """Get a ``so_map`` from an enmap (pixell format).

    Parameters
    ----------
    emap : a ndmap object
      the enmap we want to use to define the ``so_map``
    """

    new_map = so_map()
    hdulist = emap.wcs.to_fits()
    header = hdulist[0].header
    new_map.pixel = header["CTYPE1"][-3:]
    try:
        new_map.ncomp = header["NAXIS3"]
    except:
        new_map.ncomp = 1
    new_map.data = emap.copy()
    new_map.nside = None
    new_map.geometry = new_map.data.geometry[1:]
    new_map.coordinate = header["RADESYS"]
    if new_map.coordinate == "ICRS":
        new_map.coordinate = "equ"

    return new_map


def healpix2car(healpix_map, template, lmax=None):
    """Project a HEALPIX ``so_map`` into a CAR ``so_map``.

    The projection will be done in harmonic space, you can specify a lmax
    to choose a range of multipoles considered in the projection.
    If the coordinate of the map and the template differ, a rotation will be performed.

    Parameters
    ----------
    healpix_map : ``so_map`` in healpix pixellisation
      the map to be projected
    template: ``so_map`` in CAR pixellisation
      the template that will be projected onto
    lmax: integer
      the maximum multipole in the HEALPIX map to project
    """

    project = template.copy()

    if healpix_map.coordinate is None or template.coordinate is None:
        rot = None
    elif healpix_map.coordinate == template.coordinate:
        rot = None
    else:
        print(
            "will rotate from %s to %s coordinate system"
            % (healpix_map.coordinate, template.coordinate)
        )
        rot = "%s,%s" % (healpix_map.coordinate, template.coordinate)
    if lmax is None:
        lmax = 3 * healpix_map.nside - 1
    if lmax > 3 * healpix_map.nside - 1:
        print("WARNING: your lmax is too large, setting it to 3*nside-1 now")
        lmax = 3 * healpix_map.nside - 1
    project.data = reproject.enmap_from_healpix(
        healpix_map.data,
        template.data.shape,
        template.data.wcs,
        ncomp=healpix_map.ncomp,
        unit=1,
        lmax=lmax,
        rot=rot,
        first=0,
    )

    project.ncomp == healpix_map.ncomp
    if project.ncomp == 1:
        project.data = project.data[0]

    return project


def car2car(map_car, template):
    """Project a CAR map into another CAR map with different pixellisation

    Parameters
    ----------
    map : ``so_map`` in CAR pixellisation
      the map to be projected
    template: ``so_map`` in CAR pixellisation
        the template that will be projected onto
    """

    project = template.copy()
    project.data = enmap.project(map_car.data, template.data.shape, template.data.wcs)
    return project


def healpix_template(ncomp, nside, coordinate=None):
    """Create a ``so_map`` template with healpix pixellisation.

    Parameters
    ----------
    ncomp: integer
      the number of components of the map can be 1 or 3 (for T,Q,U)
    nside: integer
      the nside of the healpix map
    coordinate: string
      coordinate system of the map
    """

    temp = so_map()

    if ncomp == 3:
        temp.data = np.zeros((3, 12 * nside ** 2))
    else:
        temp.data = np.zeros((12 * nside ** 2))

    temp.pixel = "HEALPIX"
    temp.ncomp = ncomp
    temp.nside = nside
    temp.geometry = "healpix geometry"
    temp.coordinate = coordinate
    return temp


def car_template(ncomp, ra0, ra1, dec0, dec1, res):
    """Create a ``so_map`` template with CAR pixellisation in equ coordinates.

    Parameters
    ----------
    ncomp: integer
      the number of components of the map can be 1 or 3 (for T,Q,U)
    ra0,dec0,ra1,dec1: floats
      coordinates of the box in degrees
    res: float
      resolution in arcminute
    """

    if ncomp == 3:
        pre = (3,)
    else:
        pre = ()

    box = get_box(ra0, ra1, dec0, dec1)
    res = res * np.pi / (180 * 60)
    temp = so_map()
    shape, wcs = enmap.geometry(box, res=res, pre=pre)
    temp.data = enmap.zeros(shape, wcs=wcs, dtype=None)
    temp.pixel = "CAR"
    temp.nside = None
    temp.ncomp = ncomp
    temp.geometry = temp.data.geometry[1:]
    temp.coordinate = "equ"
    return temp


def full_sky_car_template(ncomp, res):
    """Create a ``so_map`` full sky template with CAR pixellisation in equ coordinates.

    Parameters
    ----------
    ncomp: integer
        the number of components of the map can be 1 or 3 (for T,Q,U)
    res: float
        resolution in arcminute
    """

    if ncomp == 3:
        pre = (3,)
    else:
        pre = ()

    res = res * np.pi / (180 * 60)
    temp = so_map()
    shape, wcs = enmap.fullsky_geometry(res=res, dims=pre)
    temp.data = enmap.zeros(shape, wcs=wcs, dtype=None)
    temp.pixel = "CAR"
    temp.nside = None
    temp.ncomp = ncomp
    temp.geometry = temp.data.geometry[1:]
    temp.coordinate = "equ"
    return temp


def white_noise(template, rms_uKarcmin_T, rms_uKarcmin_pol=None):
    """Generate a white noise realisation corresponding to the template pixellisation

    Parameters
    ----------
    template: ``so_map`` template
      the template for the white noise generalisation
    rms_uKarcmin_T: float
      the white noise temperature rms in uK.arcmin
    rms_uKarcmin_pol: float
      the white noise polarisation rms in uK.arcmin
      if None set it to sqrt(2)*rms_uKarcmin_T

    """

    noise = template.copy()
    rad_to_arcmin = 60 * 180 / np.pi
    if noise.pixel == "HEALPIX":
        nside = noise.nside
        pixArea = hp.pixelfunc.nside2pixarea(nside) * rad_to_arcmin ** 2
    if noise.pixel == "CAR":
        pixArea = noise.data.pixsizemap() * rad_to_arcmin ** 2
    if noise.ncomp == 1:
        if noise.pixel == "HEALPIX":
            size = len(noise.data)
            noise.data = np.random.randn(size) * rms_uKarcmin_T / np.sqrt(pixArea)
        if noise.pixel == "CAR":
            size = noise.data.shape
            noise.data = np.random.randn(size[0], size[1]) * rms_uKarcmin_T / np.sqrt(pixArea)
    if noise.ncomp == 3:
        if rms_uKarcmin_pol is None:
            rms_uKarcmin_pol = rms_uKarcmin_T * np.sqrt(2)
        if noise.pixel == "HEALPIX":
            size = len(noise.data[0])
            noise.data[0] = np.random.randn(size) * rms_uKarcmin_T / np.sqrt(pixArea)
            noise.data[1] = np.random.randn(size) * rms_uKarcmin_pol / np.sqrt(pixArea)
            noise.data[2] = np.random.randn(size) * rms_uKarcmin_pol / np.sqrt(pixArea)
        if noise.pixel == "CAR":
            size = noise.data[0].shape
            noise.data[0] = np.random.randn(size[0], size[1]) * rms_uKarcmin_T / np.sqrt(pixArea)
            noise.data[1] = np.random.randn(size[0], size[1]) * rms_uKarcmin_pol / np.sqrt(pixArea)
            noise.data[2] = np.random.randn(size[0], size[1]) * rms_uKarcmin_pol / np.sqrt(pixArea)

    return noise


def simulate_source_mask(binary, n_holes, hole_radius_arcmin):
    """Simulate a point source mask in a binary template

    Parameters
    ----------
    binary:  ``so_map`` binary template
      the binay map in which we generate the source mask
    n_holes: integer
      the number of masked point sources
    hole_radius_arcmin: float
      the radius of the holes
    """

    mask = binary.copy()
    if binary.pixel == "HEALPIX":
        idx = np.where(binary.data == 1)
        for i in range(n_holes):
            random_index1 = np.random.choice(idx[0])
            vec = hp.pixelfunc.pix2vec(binary.nside, random_index1)
            disc = hp.query_disc(binary.nside, vec, hole_radius_arcmin / (60.0 * 180) * np.pi)
            mask.data[disc] = 0

    if binary.pixel == "CAR":
        random_index1 = np.random.randint(0, binary.data.shape[0], size=n_holes)
        random_index2 = np.random.randint(0, binary.data.shape[1], size=n_holes)
        mask.data[random_index1, random_index2] = 0
        dist = enmap.distance_transform(mask.data)
        mask.data[dist * 60 * 180 / np.pi < hole_radius_arcmin] = 0

    return mask


def generate_source_mask(binary, coordinates, point_source_radius_arcmin):
    """Generate a point source mask in a binary template

    Parameters
    ----------
    binary:  ``so_map`` binary template
      the binay map in which we generate the source mask
    coordinates: list
      the list of point sources in DEC, RA coordinates (radians)
    point_source_radius_arcmin: float
      the radius of the point sources
    """
    mask = binary.copy()

    if mask.pixel == "HEALPIX":
        vectors = hp.ang2vec(np.pi / 2.0 - coordinates[0], 2 * np.pi - coordinates[1])
        for vec in vectors:
            disc = hp.query_disc(mask.nside, vec, point_source_radius_arcmin / (60.0 * 180) * np.pi)
            mask.data[disc] = 0

    if mask.pixel == "CAR":
        xpix, ypix = mask.data.sky2pix(coordinates).astype(int)
        mask.data[xpix, ypix] = 0
        dist = enmap.distance_transform(
            mask.data, rmax=2 * point_source_radius_arcmin * np.pi / (60.0 * 180)
        )
        mask.data[dist * 60 * 180 / np.pi < point_source_radius_arcmin] = 0

    return mask


def subtract_mono_dipole(emap, mask=None, healpix=True, bunch=24, return_values=False):
    """Subtract monopole and dipole from a ``enmap`` object.

    Parameters
    ----------
    emap: ``enmap`` or numpy array
      the map from which to compute mono/dipole values
    mask: ``enmap`` or numpy array
      a mask to put on top of the map
    healpix: bool
      flag for using HEALPIX (default) or CAR pixellisation
    bunch: int
      the bunch size (default: 24)
    return_values: bool
      Return mono/dipole values with the subtracted map (default: False)
    """
    map_cleaned = emap.copy()
    if healpix:
        map_masked = hp.ma(emap)
        if mask is not None:
            map_masked.mask = mask < 1
        mono, dipole = hp.fit_dipole(map_masked)
        npix = len(emap)
        nside = hp.npix2nside(npix)
        bunchsize = npix // bunch
        for ibunch in range(npix // bunchsize):
            ipix = np.arange(ibunch * bunchsize, (ibunch + 1) * bunchsize)
            ipix = ipix[(np.isfinite(emap.flat[ipix]))]
            x, y, z = hp.pix2vec(nside, ipix, False)
            map_cleaned.flat[ipix] -= dipole[0] * x
            map_cleaned.flat[ipix] -= dipole[1] * y
            map_cleaned.flat[ipix] -= dipole[2] * z
            map_cleaned.flat[ipix] -= mono

    else:

        def _get_xyz(dec, ra):
            x = np.cos(dec) * np.cos(ra)
            y = np.cos(dec) * np.sin(ra)
            z = np.sin(dec)
            return x, y, z

        dec, ra = emap.posmap()
        dec = dec.flatten()
        ra = ra.flatten()
        weights = emap.pixsizemap()
        weights = weights.flatten() / (4 * np.pi)

        npix = dec.size
        bunchsize = npix // bunch

        aa = np.zeros((4, 4))
        v = np.zeros(4)
        for ibunch in range(npix // bunchsize):
            ipix = np.arange(ibunch * bunchsize, (ibunch + 1) * bunchsize)
            if mask is not None:
                ipix = ipix[mask.flat[ipix] > 0]

            x, y, z = _get_xyz(dec[ipix], ra[ipix])
            w = weights[ipix]

            # aa[0, 0] += ipix.size
            aa[0, 0] += np.sum(w)
            aa[1, 0] += np.sum(x * w)
            aa[2, 0] += np.sum(y * w)
            aa[3, 0] += np.sum(z * w)
            aa[1, 1] += np.sum(x ** 2 * w)
            aa[2, 1] += np.sum(x * y * w)
            aa[3, 1] += np.sum(x * z * w)
            aa[2, 2] += np.sum(y ** 2 * w)
            aa[3, 2] += np.sum(y * z * w)
            aa[3, 3] += np.sum(z ** 2 * w)
            v[0] += np.sum(emap.flat[ipix] * w)
            v[1] += np.sum(emap.flat[ipix] * x * w)
            v[2] += np.sum(emap.flat[ipix] * y * w)
            v[3] += np.sum(emap.flat[ipix] * z * w)

        aa[0, 1] = aa[1, 0]
        aa[0, 2] = aa[2, 0]
        aa[0, 3] = aa[3, 0]
        aa[1, 2] = aa[2, 1]
        aa[1, 3] = aa[3, 1]
        aa[2, 3] = aa[3, 2]
        res = np.dot(np.linalg.inv(aa), v)
        mono = res[0]
        dipole = res[1:4]

        for ibunch in range(npix // bunchsize):
            ipix = np.arange(ibunch * bunchsize, (ibunch + 1) * bunchsize)

            x, y, z = _get_xyz(dec[ipix], ra[ipix])
            map_cleaned.flat[ipix] -= dipole[0] * x
            map_cleaned.flat[ipix] -= dipole[1] * y
            map_cleaned.flat[ipix] -= dipole[2] * z
            map_cleaned.flat[ipix] -= mono

    if return_values:
        return map_cleaned, mono, dipole
    return map_cleaned

def fourier_convolution(map_car, fourier_kernel, binary=None):

    """do a convolution in fourier space with a fourier_kernel,
    you can optionnaly use a binary to remove pathological pixels

    Parameters
    ---------
    map_car : ``so_map`` in CAR pixellisation
        the map to be convolved
    fourier_kernel: 2d array
        the convolution kernel in Fourier space
    binary:  ``so_map``
        a binary mask removing pathological pixels
        
    """
    if binary is not None:
        map_car.data *= binary.data
    ft = enmap.fft(map_car.data, normalize=True)
    ft  *= fourier_kernel
    map_car.data = enmap.ifft(ft, normalize=True).real

    return map_car



from __future__ import absolute_import, print_function

from pspy import (mcm_fortran, pspy_utils, so_config, so_cov, so_dict, so_map,
                  so_map_preprocessing, so_mcm, so_misc, so_mpi, so_spectra,
                  so_window, sph_tools, flat_tools)

from ._version import get_versions

__version__ = get_versions()['version']
del get_versions



"""
A module to handle base configuration
"""

from __future__ import absolute_import, print_function

import argparse
import configparser
import os

PSPIPE_ROOT = os.path.join(os.path.dirname(os.path.abspath(__file__)), "..")
DEFAULT_OUTPUT_DIR = os.path.join(PSPIPE_ROOT, "output")
DEFAULT_DATA_DIR = os.path.join(PSPIPE_ROOT, "data")
DEFAULT_CONFIG_DIR = os.path.join(PSPIPE_ROOT, "configs")

argparser = argparse.ArgumentParser(
    prog="pspipe",
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    conflict_handler="resolve",
)

configparser = configparser.ConfigParser()
configparser.optionxform = str


def get_output_dir():
    """ Return default output directory """
    return DEFAULT_OUTPUT_DIR


def get_data_dir():
    """ Return default data directory """
    return DEFAULT_DATA_DIR


def load_config(config_file):
    global configparser

    path_to_config = os.path.join(DEFAULT_CONFIG_DIR, config_file)
    configparser.read(path_to_config)



import numpy as np, pylab as plt
from pspy import flat_tools, pspy_utils
from pixell import enmap, utils

def build_std_filter(shape, wcs, vk_mask, hk_mask, dtype=np.float64):
    ly, lx  = enmap.laxes(shape, wcs, method="auto")
    ly = ly.astype(dtype)
    lx = lx.astype(dtype)
    filter = enmap.ones(shape[-2:], wcs, dtype)
    if vk_mask is not None:
        id_vk = np.where((lx > vk_mask[0]) & (lx < vk_mask[1]))
        filter[:, id_vk] = 0
    if hk_mask is not None:
        id_hk = np.where((ly > hk_mask[0]) & (ly < hk_mask[1]))
        filter[id_hk, :] = 0
    return filter

def apply_std_filter(imap, filter):
    filtered_map = imap.copy()
    if imap.ncomp == 1:
        filtered_map.data = enmap.ifft(filter*enmap.fft(imap.data)).real
    else:
        for i in range(imap.ncomp):
            filtered_map.data[i] = enmap.ifft(filter*enmap.fft(imap.data[i])).real

    return filtered_map

def build_sigurd_filter(shape, wcs, lbounds, dtype=np.float64):
    ly, lx  = enmap.laxes(shape, wcs, method="auto")
    ly = ly.astype(dtype)
    lx = lx.astype(dtype)
    lbounds = np.asarray(lbounds)
    if lbounds.ndim < 2:
        lbounds = np.broadcast_to(lbounds, (1,2))
    if lbounds.ndim > 2 or lbounds.shape[-1] != 2:
        raise ValueError("lbounds must be [:,{ly,lx}]")
    filter = enmap.ones(shape[-2:], wcs, dtype)
    # Apply the filters
    for i , (ycut, xcut) in enumerate(lbounds):
        filter *= 1-(np.exp(-0.5*(ly/ycut)**2)[:,None]*np.exp(-0.5*(lx/xcut)**2)[None,:])
    return filter

def apply_sigurd_filter(imap, ivar, filter, tol=1e-4, ref=0.9):
    """Filter enmap imap with the given 2d fourier filter while
    weithing spatially with ivar"""
    
    filtered_map = imap.copy()
    rhs    = enmap.ifft((1-filter)*enmap.fft(imap.data*ivar)).real
    div    = enmap.ifft((1-filter)*enmap.fft(ivar)).real
    # Avoid division by very low values
    div    = np.maximum(div, np.percentile(ivar[::10,::10],ref*100)*tol)
    filtered_map.data = imap.data - rhs/div
    return filtered_map

def analytical_tf(map, filter, binning_file, lmax):
    lmap  = map.data.modlmap() 
    twod_index = lmap.copy()
    bin_lo, bin_hi, bin_c, bin_size = pspy_utils.read_binning_file(binning_file, lmax)
    nbins = len(bin_lo)
    tf = np.zeros(nbins)
    for ii in range(nbins):
        id = np.where((lmap >= bin_lo[ii]) & (lmap <= bin_hi[ii]))
        twod_index *= 0
        twod_index[id] = 1
        bin_area = np.sum(twod_index)
        masked_bin_area= np.sum(twod_index * filter)
        tf[ii] = masked_bin_area / bin_area

    return bin_c, tf



#def analytical_tf_old(map, binning_file, lmax, vk_mask=None, hk_mask=None):
#    import time
#    t = time.time()
#    ft = flat_tools.fft_from_so_map(map)
#    ft.create_kspace_mask(vertical_stripe=vk_mask, horizontal_stripe=hk_mask)
#    lmap = ft.lmap
#    kmask = ft.kmask
#    twod_index = lmap.copy()

#    bin_lo, bin_hi, bin_c, bin_size = pspy_utils.read_binning_file(binning_file, lmax)
#    nbins = len(bin_lo)
#    print(lmap.shape)
#    tf = np.zeros(nbins)
#    print("init tf", time.time()-t)
#    for ii in range(nbins):
#        id = np.where( (lmap >= bin_lo[ii]) & (lmap <= bin_hi[ii]))
#        twod_index *= 0
#        twod_index[id] = 1
#        bin_area = np.sum(twod_index)
#        masked_bin_area= np.sum(twod_index * kmask)
#        tf[ii] = masked_bin_area / bin_area
#    print("loop", time.time()-t)

#    return bin_c, tf


#def kspace_filter_old(map, vk_mask=None, hk_mask=None, window=None, normalize=False):
#    filtered_map = map.copy()

 #   ft = flat_tools.fft_from_so_map(map, normalize=normalize)
 #   if window is not None:
 #       ft = flat_tools.get_ffts(map, window, normalize=normalize)

 #   if vk_mask is not None:
 #       id_vk = np.where((ft.lx > vk_mask[0]) & (ft.lx < vk_mask[1]))
 #   if hk_mask is not None:
 #       id_hk = np.where((ft.ly > hk_mask[0]) & (ft.ly < hk_mask[1]))

 #   if map.ncomp == 1:
 #       if vk_mask is not None:
 #           ft.kmap[: , id_vk] = 0.
 #       if hk_mask is not None:
 #           ft.kmap[id_hk , :] = 0.
    
 #   if map.ncomp == 3:
 #      for i in range(3):
 #           if vk_mask is not None:
 #               ft.kmap[i, : , id_vk] = 0.
 #           if hk_mask is not None:
 #               ft.kmap[i, id_hk , :] = 0.

 #   filtered_map.data[:] = ft.map_from_fft(normalize=normalize)
 #   return filtered_map



"""
Tools for analytical covariance matrix estimation. For more details on computation of the matrix see
https://pspy.readthedocs.io/en/latest/scientific_doc.pdf.
"""
import healpy as hp
import numpy as np
from pixell import colorize, enmap, enplot
from pspy import pspy_utils, so_mcm, sph_tools
from pspy.cov_fortran.cov_fortran import cov_compute as cov_fortran
from pspy.mcm_fortran.mcm_fortran import mcm_compute as mcm_fortran


def cov_coupling_spin0(win, lmax, niter=3, save_file=None, l_exact=None, l_band=None, l_toep=None):
    """compute the coupling kernels corresponding to the T only covariance matrix
        
   Parameters
    ----------

    win: so_map or dictionnary of so_map
      the window functions, can be a so_map or a dictionnary containing so_map
      if the later, the entry of the dictionnary should be Ta,Tb,Tc,Td.
    lmax: integer
      the maximum multipole to consider
    niter: int
      the number of iteration performed while computing the alm
    save_file: string
      the name of the file in which the coupling kernel will be saved (npy format)
    l_toep: int
      parameter for the toeplitz approximation (see arXiv:2010.14344)
    l_band: int
      parameter for the toeplitz approximation (see arXiv:2010.14344)
    l_exact: int
      parameter for the toeplitz approximation (see arXiv:2010.14344)
    """

    coupling_dict = {}
    if l_toep is None: l_toep = lmax
    if l_band is None: l_band = lmax
    if l_exact is None: l_exact = lmax

    if type(win) is not dict:
        sq_win = win.copy()
        sq_win.data *= sq_win.data
        alm = sph_tools.map2alm(sq_win, niter=niter, lmax=lmax)
        wcl = hp.alm2cl(alm)
        l = np.arange(len(wcl))
        wcl *= (2 * l + 1) / (4 * np.pi)
        coupling = np.zeros((lmax, lmax))
        

        mcm_fortran.calc_coupling_spin0(wcl,
                                       l_exact,
                                       l_band,
                                       l_toep,
                                       coupling.T)

        
        if l_toep < lmax:
            # For toepliz fill the matrix
            coupling = so_mcm.format_toepliz_fortran(coupling, l_toep, lmax)

        mcm_fortran.fill_upper(coupling.T)

        coupling_dict["TaTcTbTd"] = coupling[:lmax - 2, :lmax - 2]
        coupling_dict["TaTdTbTc"] = coupling[:lmax - 2, :lmax - 2]
        
    else:
        wcl = {}
        for s in ["TaTcTbTd", "TaTdTbTc"]:
            n0, n1, n2, n3 = [s[i * 2:(i + 1) * 2] for i in range(4)]
            sq_win_n0n1 = win[n0].copy()
            sq_win_n0n1.data *= win[n1].data
            sq_win_n2n3 = win[n2].copy()
            sq_win_n2n3.data *= win[n3].data
            alm_n0n1 = sph_tools.map2alm(sq_win_n0n1, niter=niter, lmax=lmax)
            alm_n2n3 = sph_tools.map2alm(sq_win_n2n3, niter=niter, lmax=lmax)
            wcl[n0+n1+n2+n3] = hp.alm2cl(alm_n0n1, alm_n2n3)
            l = np.arange(len(wcl[n0 + n1 + n2 + n3]))
            wcl[n0+n1+n2+n3] *= (2 * l + 1) / (4 * np.pi)

        coupling = np.zeros((2, lmax, lmax))
        
        cov_fortran.calc_cov_spin0(wcl["TaTcTbTd"], wcl["TaTdTbTc"], l_exact, l_band, l_toep,  coupling.T)
        
        for id_cov, name in enumerate(["TaTcTbTd", "TaTdTbTc"]):
            if l_toep < lmax:
                coupling[id_cov] = so_mcm.format_toepliz_fortran(coupling[id_cov], l_toep, lmax)
            mcm_fortran.fill_upper(coupling[id_cov].T)

            coupling_dict[name] = coupling[id_cov][:lmax - 2, :lmax - 2]

        
    if save_file is not None:
        np.save("%s.npy"%save_file, coupling)

    return coupling_dict


def cov_coupling_spin0and2_simple(win, lmax, niter=3, save_file=None, planck=False, l_exact=None, l_band=None, l_toep=None):
    """Compute the coupling kernels corresponding to the T and E covariance matrix

    Parameters
    ----------

    win: so_map or dictionnary of so_map
      the window functions, can be a so_map or a dictionnary containing so_map
      if the later, the entry of the dictionnary should be Ta,Tb,Tc,Td,Pa,Pb,Pc,Pd
    lmax: integer
      the maximum multipole to consider
    niter: int
      the number of iteration performed while computing the alm
    save_file: string
      the name of the file in which the coupling kernel will be saved (npy format)
    l_toep: int
      parameter for the toeplitz approximation (see arXiv:2010.14344)
    l_band: int
      parameter for the toeplitz approximation (see arXiv:2010.14344)
    l_exact: int
      parameter for the toeplitz approximation (see arXiv:2010.14344)


    """


    win_list = ["TaTcTbTd", "TaTdTbTc", "PaPcPbPd", "PaPdPbPc",
                "TaTcPbPd", "TaPdPbTc", "PaPcTbTd", "PaTdTbPc",
                "TaPcTbPd", "TaPdTbPc", "TaTcTbPd", "TaPdTbTc",
                "TaPcTbTd", "TaTdTbPc", "TaPcPbTd", "TaTdPbPc",
                "TaPcPbPd", "TaPdPbPc", "PaPcTbPd", "PaPdTbPc",
                "TaTcPbTd", "TaTdPbTc", "PaTcTbTd", "PaTdTbTc",
                "PaTcPbTd", "PaTdPbTc", "PaTcTbPd", "PaPdTbTc",
                "PaTcPbPd", "PaPdPbTc", "PaPcPbTd", "PaTdPbPc"]
                
    coupling_dict = {}
    
    if l_toep is None: l_toep = lmax
    if l_band is None: l_band = lmax
    if l_exact is None: l_exact = lmax

    if type(win) is not dict:
        sq_win = win.copy()
        sq_win.data *= sq_win.data
        alm = sph_tools.map2alm(sq_win, niter=niter, lmax=lmax)
        wcl = hp.alm2cl(alm)
        l = np.arange(len(wcl))
        wcl *= (2 * l + 1) / (4 * np.pi)
        coupling = np.zeros((1, lmax, lmax))
    

        mcm_fortran.calc_coupling_spin0(wcl,
                                       l_exact,
                                       l_band,
                                       l_toep,
                                       coupling[0].T)

        
        if l_toep < lmax:
            # For toepliz fill the matrix
            coupling[0] = so_mcm.format_toepliz_fortran(coupling[0], l_toep, maxl)

        indexlist=np.zeros(32, dtype=np.int8)

        for name,index in zip(win_list, indexlist):
            coupling_dict[name] = coupling[index] + coupling[index].T - np.diag(np.diag(coupling[index]))
            coupling_dict[name] = coupling_dict[name][:lmax - 2, :lmax - 2]
    else:
        wcl={}
        
        for s in win_list:

            n0, n1, n2, n3 = [s[i * 2:(i + 1) * 2] for i in range(4)]

            sq_win_n0n1 = win[n0].copy()
            sq_win_n0n1.data *= win[n1].data
            sq_win_n2n3 = win[n2].copy()
            sq_win_n2n3.data *= win[n3].data

            alm_n0n1 = sph_tools.map2alm(sq_win_n0n1, niter=niter, lmax=lmax)
            alm_n2n3 = sph_tools.map2alm(sq_win_n2n3, niter=niter, lmax=lmax)

            wcl[n0+n1+n2+n3] = hp.alm2cl(alm_n0n1, alm_n2n3)
            l = np.arange(len(wcl[n0+n1+n2+n3]))
            wcl[n0+n1+n2+n3] *= (2 * l + 1) / (4 * np.pi)

        coupling = np.zeros((32, lmax, lmax))

        if planck==True:
            doPlanck = 1
        else:
            doPlanck = 0
            
            
        cov_fortran.calc_cov_spin0and2_simple(wcl["TaTcTbTd"], wcl["TaTdTbTc"], wcl["PaPcPbPd"], wcl["PaPdPbPc"],
                                              wcl["TaTcPbPd"], wcl["TaPdPbTc"], wcl["PaPcTbTd"], wcl["PaTdTbPc"],
                                              wcl["TaPcTbPd"], wcl["TaPdTbPc"], wcl["TaTcTbPd"], wcl["TaPdTbTc"],
                                              wcl["TaPcTbTd"], wcl["TaTdTbPc"], wcl["TaPcPbTd"], wcl["TaTdPbPc"],
                                              wcl["TaPcPbPd"], wcl["TaPdPbPc"], wcl["PaPcTbPd"], wcl["PaPdTbPc"],
                                              wcl["TaTcPbTd"], wcl["TaTdPbTc"], wcl["PaTcTbTd"], wcl["PaTdTbTc"],
                                              wcl["PaTcPbTd"], wcl["PaTdPbTc"], wcl["PaTcTbPd"], wcl["PaPdTbTc"],
                                              wcl["PaTcPbPd"], wcl["PaPdPbTc"], wcl["PaPcPbTd"], wcl["PaTdPbPc"],
                                              doPlanck, l_exact, l_band, l_toep, coupling.T)


        indexlist = np.arange(32)
        for name, index in zip(win_list, indexlist):
            if l_toep < lmax:
                coupling[index] = so_mcm.format_toepliz(coupling[index], l_toep, lmax)
            mcm_fortran.fill_upper(coupling[index].T)
            coupling_dict[name] = coupling[index][:lmax - 2, :lmax - 2]

    if save_file is not None:
        np.save("%s.npy"%save_file, coupling)

    return coupling_dict
    
    
def fast_cov_coupling_spin0and2(sq_win_alms_dir,
                                id_element,
                                lmax,
                                l_exact=None,
                                l_band=None,
                                l_toep=None):
                                
    """Compute the coupling kernels corresponding to the T and E covariance matrix
    This routine assume that you have already precomputed the alms of the square of the windows
    and that the window in temperature and polarisation are the same.
    Both conditions lead to a very important speed up, that explains the name of the routine

    Parameters
    ----------

    sq_win_alms_dir: string
        the folder with precomputed alms corresponding to the sq of the window function
        this is what enters the analytic computation
    id_element : list
        a list of the form [a,b,c,d] where a = dr6_pa4_090, etc, this identify which pair of power spectrum we want the covariance of
    lmax: integer
      the maximum multipole to consider
    l_toep: int
      parameter for the toeplitz approximation (see arXiv:2010.14344)
    l_band: int
      parameter for the toeplitz approximation (see arXiv:2010.14344)
    l_exact: int
      parameter for the toeplitz approximation (see arXiv:2010.14344)

    """

    na, nb, nc, nd = id_element

    if l_toep is None: l_toep = lmax
    if l_band is None: l_band = lmax
    if l_exact is None: l_exact = lmax
    
    def try_alm_order(dir, a, b):
        try:
            alm = np.load("%s/alms_%sx%s.npy"  % (dir, a, b))
        except:
            alm = np.load("%s/alms_%sx%s.npy"  % (dir, b, a))
        return alm
    
    alm_TaTc = try_alm_order(sq_win_alms_dir, na, nc)
    alm_TbTd = try_alm_order(sq_win_alms_dir, nb, nd)
    alm_TaTd = try_alm_order(sq_win_alms_dir, na, nd)
    alm_TbTc = try_alm_order(sq_win_alms_dir, nb, nc)

    wcl = {}
    wcl["TaTcTbTd"] = hp.alm2cl(alm_TaTc, alm_TbTd)
    wcl["TaTdTbTc"] = hp.alm2cl(alm_TaTd, alm_TbTc)

    l = np.arange(len(wcl["TaTcTbTd"]))
    wcl["TaTcTbTd"] *= (2 * l + 1) / (4 * np.pi)
    wcl["TaTdTbTc"] *= (2 * l + 1) / (4 * np.pi)

    coupling = np.zeros((2, lmax, lmax))
 
    cov_fortran.calc_cov_spin0(wcl["TaTcTbTd"], wcl["TaTdTbTc"], l_exact, l_band, l_toep,  coupling.T)
 
    coupling_dict = {}

    for id_cov, name in enumerate(["TaTcTbTd", "TaTdTbTc"]):
        if l_toep < lmax:
            coupling[id_cov] = so_mcm.format_toepliz_fortran(coupling[id_cov], l_toep, lmax)
        mcm_fortran.fill_upper(coupling[id_cov].T)
        coupling_dict[name] = coupling[id_cov]

    list1 = ["TaTcTbTd", "PaPcPbPd", "TaTcPbPd", "PaPcTbTd",
             "TaPcTbPd", "TaTcTbPd", "TaPcTbTd", "TaPcPbTd",
             "TaPcPbPd", "PaPcTbPd", "TaTcPbTd", "PaTcTbTd",
             "PaTcPbTd", "PaTcTbPd", "PaTcPbPd", "PaPcPbTd"]
        
    list2 = ["TaTdTbTc", "PaPdPbPc", "TaPdPbTc", "PaTdTbPc",
             "TaPdTbPc", "TaPdTbTc", "TaTdTbPc", "TaTdPbPc",
             "TaPdPbPc", "PaPdTbPc", "TaTdPbTc", "PaTdTbTc",
             "PaTdPbTc", "PaPdTbTc", "PaPdPbTc", "PaTdPbPc"]
 
    for id1 in list1:
        coupling_dict[id1] = coupling_dict["TaTcTbTd"][:lmax - 2, :lmax - 2]
    for id2 in list2:
        coupling_dict[id2] = coupling_dict["TaTdTbTc"][:lmax - 2, :lmax - 2]
        
    return coupling_dict



def read_coupling(file, spectra=None):
    """Read a precomputed coupling kernels
    the code use the size of the array to infer what type of survey it corresponds to

    Parameters
    ----------

    file: string
      the name of the npy file
    """

    coupling = np.load("%s.npy"%file)
    coupling_dict = {}
    if spectra is None:
        win_list = ["TaTcTbTd", "TaTdTbTc"]
        if coupling.shape[0] == 2:
            indexlist = [0,1]
        elif coupling.shape[0] == 1:
            indexlist = [0,0]
    else:
        win_list = ["TaTcTbTd", "TaTdTbTc", "PaPcPbPd", "PaPdPbPc",
                    "TaTcPbPd", "TaPdPbTc", "PaPcTbTd", "PaTdTbPc",
                    "TaPcTbPd", "TaPdTbPc", "TaTcTbPd", "TaPdTbTc",
                    "TaPcTbTd", "TaTdTbPc", "TaPcPbTd", "TaTdPbPc",
                    "TaPcPbPd", "TaPdPbPc", "PaPcTbPd", "PaPdTbPc",
                    "TaTcPbTd", "TaTdPbTc", "PaTcTbTd", "PaTdTbTc",
                    "PaTcPbTd", "PaTdPbTc", "PaTcTbPd", "PaPdTbTc",
                    "PaTcPbPd", "PaPdPbTc", "PaPcPbTd", "PaTdPbPc"]
                    
        if coupling.shape[0] == 32:
            indexlist = np.arange(32)
        elif coupling.shape[0] == 1:
            indexlist=np.zeros(32, dtype=np.int8)
        
    for name, index in zip(win_list, indexlist):
        coupling_dict[name] = coupling[index]

    return coupling_dict


def symmetrize(Clth, mode="arithm"):
    """Take a power spectrum Cl and return a symmetric array C_l1l2=f(Cl)

    Parameters
    ----------
    Clth: 1d array
      the power spectrum to be made symmetric
    mode : string
      geometric or arithmetic mean
      if geo return C_l1l2 = sqrt( |Cl1 Cl2 |)
      if arithm return C_l1l2 = (Cl1 + Cl2)/2
    """
    
    if mode == "geo":
        return np.sqrt(np.abs(np.outer(Clth, Clth)))
    if mode == "arithm":
        return np.add.outer(Clth, Clth) / 2

def bin_mat(mat, binning_file, lmax, speclist=["TT"]):
    """Take a matrix and bin it Mbb'= Pbl Pb'l' Mll' with  Pbl =1/Nb sum_(l in b)

    Parameters
    ----------
    binning_file: data file
      a binning file with format bin low, bin high, bin mean
    lmax: int
      the maximum multipole to consider
    """

    bin_lo, bin_hi, bin_c, bin_size = pspy_utils.read_binning_file(binning_file, lmax)

    n_bins = len(bin_hi)
    n_spec = len(speclist)
    n_ell = int(mat.shape[0] / n_spec)
   

    coupling_b = np.zeros((n_spec * n_bins, n_spec * n_bins))

    for i, s1 in enumerate(speclist):
        for j, s2 in enumerate(speclist):
        
            block = mat[i * n_ell : (i + 1) * n_ell,  j * n_ell : (j + 1) * n_ell]
            binned_block = np.zeros((n_bins, n_bins))
            mcm_fortran.bin_mcm(block.T, bin_lo, bin_hi, bin_size, binned_block.T, 0)
            binned_block /= bin_size
            coupling_b[i * n_bins : (i + 1) * n_bins,  j * n_bins : (j + 1) * n_bins] = binned_block

    return coupling_b

def cov_spin0(Clth_dict, coupling_dict, binning_file, lmax, mbb_inv_ab, mbb_inv_cd, binned_mcm=True):
    """From the two point functions and the coupling kernel construct the spin0 analytical covariance matrix of <(C_ab- Clth)(C_cd-Clth)>

    Parameters
    ----------

    Clth_dict: dictionnary
      A dictionnary of theoretical power spectrum (auto and cross) for the different split combinaison ('TaTb' etc)
    coupling_dict: dictionnary
      a dictionnary containing the coupling kernel
    binning_file: data file
      a binning file with format bin low, bin high, bin mean
    lmax: int
      the maximum multipole to consider
    mbb_inv_ab: 2d array
      the inverse mode coupling matrix for the 'TaTb' power spectrum
    mbb_inv_cd: 2d array
      the inverse mode coupling matrix for the 'TcTd' power spectrum
    binned_mcm: boolean
      specify if the mode coupling matrices are binned or not

    """

    cov = symmetrize(Clth_dict["TaTc"]) * symmetrize(Clth_dict["TbTd"]) * coupling_dict["TaTcTbTd"]
    cov += symmetrize(Clth_dict["TaTd"]) * symmetrize(Clth_dict["TbTc"]) * coupling_dict["TaTdTbTc"]

    if binned_mcm == True:
        analytic_cov = bin_mat(cov, binning_file, lmax)
        analytic_cov = np.dot(np.dot(mbb_inv_ab, analytic_cov), mbb_inv_cd.T)
    else:
        full_analytic_cov = np.dot(np.dot(mbb_inv_ab, cov), mbb_inv_cd.T)
        analytic_cov = bin_mat(full_analytic_cov, binning_file, lmax)
        
    return analytic_cov

def cov_spin0and2(Clth_dict, coupling_dict, binning_file, lmax, mbb_inv_ab, mbb_inv_cd, binned_mcm=True):
    """From the two point functions and the coupling kernel construct the T and E analytical covariance matrix of <(C_ab- Clth)(C_cd-Clth)>

    Parameters
    ----------

    Clth_dict: dictionnary
        A dictionnary of theoretical power spectrum (auto and cross) for the different split combinaison ('XaYb' etc)
    coupling_dict: dictionnary
        a dictionnary containing the coupling kernel
    binning_file: data file
        a binning file with format bin low, bin high, bin mean
    lmax: int
        the maximum multipole to consider
    mbb_inv_ab: 2d array
        the inverse mode coupling matrix for the 'XaYb' power spectrum
    mbb_inv_cd: 2d array
        the inverse mode coupling matrix for the 'XcYd' power spectrum
    binned_mcm: boolean
      specify if the mode coupling matrices are binned or not

    """

    bin_lo, bin_hi, bin_c, bin_size = pspy_utils.read_binning_file(binning_file, lmax)
    n_ell = Clth_dict["TaTb"].shape[0]
        
    full_analytic_cov = np.zeros((4 * n_ell, 4 * n_ell))
    
    speclist = ["TT", "TE", "ET", "EE"]
    for i, (W, X) in enumerate(speclist):
        for j, (Y, Z) in enumerate(speclist):
            if i > j : continue
            id0 = W + "a" + Y + "c"
            id1 = X + "b" + Z + "d"
            id2 = W + "a" + Z + "d"
            id3 = X + "b" + Y + "c"

            M = symmetrize(Clth_dict[id0]) * symmetrize(Clth_dict[id1]) * coupling_dict[id0.replace("E","P") + id1.replace("E","P")]
            M += symmetrize(Clth_dict[id2]) * symmetrize(Clth_dict[id3]) * coupling_dict[id2.replace("E","P") + id3.replace("E","P")]
            
            full_analytic_cov[i * n_ell:(i + 1) * n_ell, j * n_ell:(j + 1) * n_ell] = M
    
    full_analytic_cov = np.triu(full_analytic_cov) + np.tril(full_analytic_cov.T, -1)
    
    mbb_inv_ab = extract_TTTEEE_mbb(mbb_inv_ab)
    mbb_inv_cd = extract_TTTEEE_mbb(mbb_inv_cd)

    if binned_mcm == True:
        analytic_cov = bin_mat(full_analytic_cov, binning_file, lmax, speclist=speclist)
        analytic_cov = np.dot(np.dot(mbb_inv_ab, analytic_cov), mbb_inv_cd.T)
    else:
        full_analytic_cov = np.dot(np.dot(mbb_inv_ab, full_analytic_cov), mbb_inv_cd.T)
        analytic_cov = bin_mat(full_analytic_cov, binning_file, lmax, speclist=speclist)

    return analytic_cov

def generalized_cov_spin0and2(coupling_dict, id_element, ns, ps_all, nl_all, lmax, binning_file, mbb_inv_ab, mbb_inv_cd, binned_mcm=True, return_full_cov=False):

    """
    This routine deserves some explanation
    We want to compute the covariance between two power spectra
    C1 = Wa * Xb, C2 =  Yc * Zd
    Here W, X, Y, Z can be either T or E and a,b,c,d will be an index
    corresponding to the survey and array we consider so for example a = s17&pa5_150 or a = dr6&pa4_090
    The formula for the analytic covariance of C1, C2 is given by
    Cov( Wa * Xb,  Yc * Zd) = < Wa Yc> <Xb Zd>  + < Wa Zd> <Xb Yc> (this is just from the wick theorem)
    In practice we need to include the effect of the mask (so we have to introduce the coupling dict D)
    and we need to take into account that we use as spectra an average of cross power spectra, that is why we use the chi function
    (and what make it different from cov_spin0and2)
    Cov( Wa * Xb,  Yc * Zd) = D(Wa*Yc,Xb Zd) chi(Wa,Yc,Xb Zd) +  D(Wa*Zd,Xb*Yc) chi(Wa,Zd,Xb,Yc)

    Parameters
    ----------
    coupling_dict : dictionnary
        a dictionnary that countains the coupling terms arising from the window functions
    id_element : list
        a list of the form [a,b,c,d] where a = dr6_pa4_090, etc, this identify which pair of power spectrum we want the covariance of
    ns: dict
        this dictionnary contains the number of split we consider for each of the survey
    ps_all: dict
        this dict contains the theoretical best power spectra, convolve with the beam for example
        ps["dr6&pa5_150", "dr6&pa4_150", "TT"] = bl_dr6_pa5_150 * bl_dr6_pa4_150 * (Dl^{CMB}_TT + fg_TT)
    nl_all: dict
        this dict contains the estimated noise power spectra, note that it correspond to the noise power spectrum per split
        e.g nl["dr6&pa5_150", "dr6&pa4_150", "TT"]
    binning_file:
        a binning file with three columns bin low, bin high, bin mean
    mbb_inv_ab and mbb_inv_cd:
        the inverse mode coupling matrices corresponding to the C1 = Wa * Xb and C2 =  Yc * Zd power spectra
    binned_mcm: boolean
        specify if the mode coupling matrices are binned or not
    return_full_cov: boolean
        an option to return the lbyl cov (if binned_mcm=False)
        mostly used for debugging

    """

    na, nb, nc, nd = id_element

    n_ell = coupling_dict["TaTcTbTd"].shape[0]
    bin_lo, bin_hi, bin_c, bin_size = pspy_utils.read_binning_file(binning_file, n_ell)
    full_analytic_cov = np.zeros((4 * n_ell, 4 * n_ell))

    speclist = ["TT", "TE", "ET", "EE"]
    for i, (W, X) in enumerate(speclist):
        for j, (Y, Z) in enumerate(speclist):
    
            id0 = W + "a" + Y + "c"
            id1 = X + "b" + Z + "d"
            id2 = W + "a" + Z + "d"
            id3 = X + "b" + Y + "c"
        
            M = coupling_dict[id0.replace("E","P") + id1.replace("E","P")] * chi(na, nc, nb, nd, ns, ps_all, nl_all, W + Y + X + Z)
            M += coupling_dict[id2.replace("E","P") + id3.replace("E","P")] * chi(na, nd, nb, nc, ns, ps_all, nl_all, W + Z + X + Y)
            full_analytic_cov[i * n_ell: (i + 1) * n_ell, j * n_ell: (j + 1) * n_ell] = M

    mbb_inv_ab = extract_TTTEEE_mbb(mbb_inv_ab)
    mbb_inv_cd = extract_TTTEEE_mbb(mbb_inv_cd)

    if binned_mcm == True:
        analytic_cov = bin_mat(full_analytic_cov, binning_file, lmax, speclist=speclist)
        analytic_cov = np.dot(np.dot(mbb_inv_ab, analytic_cov), mbb_inv_cd.T)
    else:
        full_analytic_cov = np.dot(np.dot(mbb_inv_ab, full_analytic_cov), mbb_inv_cd.T)
        if return_full_cov == True:
            return full_analytic_cov

        analytic_cov = bin_mat(full_analytic_cov, binning_file, lmax, speclist=speclist)

    return analytic_cov


def covariance_element_beam(id_element, ps_all, norm_beam_cov, binning_file, lmax):
    """
    This routine compute the contribution from beam errors to the analytical covariance of the power spectra
    We want to compute the beam covariance between the two spectra
    C1 = Wa * Xb, C2 =  Yc * Zd
    Here W, X, Y, Z can be either T or E and a,b,c,d will be an index
    corresponding to the survey and array we consider so for example a = dr6&pa5_150 or a = dr6&pa4_090
    The formula for the analytic covariance of C1, C2 is given by
    let's denote the normalised beam covariance <BB>_ac = < delta B_a delta B_c >/np.outer(B_a, B_c)

    Cov(Wa * Xb,  Yc * Zd) = Dl^{WaXb} Dl^{YcZd} ( <BB>_ac + <BB>_ad + <BB>_bc + <BB>_bd )
   
    Parameters
    ----------
    id_element : list
        a list of the form [a,b,c,d] where a = dr6_pa4_090, etc, this identify which pair of power spectrum we want the covariance of
    ps_all: dict
        this dict contains the theoretical best power spectra, convolve with the beam for example
        ps["dr6&pa5_150", "dr6&pa4_150", "TT"] =  (Dl^{CMB}_TT + fg_TT)
    norm_beam_cov: dict
        this dict contains the normalized beam covariance for each survey and array
    binning_file: str
        a binning file with three columns bin low, bin high, bin mean
    lmax: int
        the maximum multipole to consider
    """
    na, nb, nc, nd = id_element

    sv_alpha, ar_alpha = na.split("&")
    sv_beta, ar_beta = nb.split("&")
    sv_gamma, ar_gamma = nc.split("&")
    sv_eta, ar_eta = nd.split("&")

    bin_lo, bin_hi, bin_c, bin_size = pspy_utils.read_binning_file(binning_file, lmax)
    nbins = len(bin_hi)

    speclist = ["TT", "TE", "ET", "EE"]

    nspec = len(speclist)
    analytic_cov_from_beam = np.zeros((nspec * nbins, nspec * nbins))
    for i, spec1 in enumerate(speclist):
        for j, spec2 in enumerate(speclist):

            M =  (delta2(na, nc) + delta2(na, nd)) * norm_beam_cov[sv_alpha, ar_alpha]
            M += (delta2(nb, nc) + delta2(nb, nd)) * norm_beam_cov[sv_beta, ar_beta]
            M *=  np.outer(ps_all[na, nb, spec1], ps_all[nc, nd, spec2])
        
            analytic_cov_from_beam[i * nbins: (i + 1) * nbins, j * nbins: (j + 1) * nbins] = bin_mat(M, binning_file, lmax)

    return analytic_cov_from_beam


def extract_TTTEEE_mbb(mbb_inv):
    """The mode coupling marix is computed for T,E,B but for now we only construct analytical covariance matrix for T and E
    The B modes is complex with important E->B leakage, this routine extract the T and E part of the mode coupling matrix

    Parameters
    ----------

    mbb_inv: 2d array
      the inverse spin0 and 2 mode coupling matrix
    """

    mbb_inv_array = so_mcm.coupling_dict_to_array(mbb_inv)
    mbb_array = np.linalg.inv(mbb_inv_array)
    nbins = int(mbb_array.shape[0] / 9)
    
    mbb_array_select = np.zeros((4*nbins, 4*nbins))
    # TT
    mbb_array_select[0*nbins:1*nbins, 0*nbins:1*nbins] = mbb_array[0*nbins:1*nbins, 0*nbins:1*nbins]
    # TE
    mbb_array_select[1*nbins:2*nbins, 1*nbins:2*nbins] = mbb_array[1*nbins:2*nbins, 1*nbins:2*nbins]
    # ET
    mbb_array_select[2*nbins:3*nbins, 2*nbins:3*nbins] = mbb_array[3*nbins:4*nbins, 3*nbins:4*nbins]
    # EE
    mbb_array_select[3*nbins:4*nbins, 3*nbins:4*nbins] = mbb_array[5*nbins:6*nbins, 5*nbins:6*nbins]

    mbb_inv_array = np.linalg.inv(mbb_array_select)
    
    return mbb_inv_array

def extract_EEEBBB_mbb(mbb_inv):
    """this routine extract the E and B part of the mode coupling matrix

    Parameters
    ----------

    mbb_inv: 2d array
        the inverse spin0 and 2 mode coupling matrix
    """

    mbb_inv_array = so_mcm.coupling_dict_to_array(mbb_inv)
    mbb_array = np.linalg.inv(mbb_inv_array)
    nbins = int(mbb_array.shape[0] / 9)

    mbb_array_select = np.zeros((4*nbins, 4*nbins))
    # EE
    mbb_array_select[0*nbins:1*nbins, 0*nbins:1*nbins] = mbb_array[5*nbins:6*nbins, 5*nbins:6*nbins]
    # EB
    mbb_array_select[1*nbins:2*nbins, 1*nbins:2*nbins] = mbb_array[6*nbins:7*nbins, 6*nbins:7*nbins]
    # BE
    mbb_array_select[2*nbins:3*nbins, 2*nbins:3*nbins] = mbb_array[7*nbins:8*nbins, 7*nbins:8*nbins]
    # BB
    mbb_array_select[3*nbins:4*nbins, 3*nbins:4*nbins] = mbb_array[8*nbins:9*nbins, 8*nbins:9*nbins]
    # EE-BB
    mbb_array_select[0*nbins:1*nbins, 3*nbins:4*nbins] = mbb_array[5*nbins:6*nbins, 8*nbins:9*nbins]
    # BB-EE
    mbb_array_select[3*nbins:4*nbins, 0*nbins:1*nbins] = mbb_array[8*nbins:9*nbins, 5*nbins:6*nbins]

    mbb_inv_array = np.linalg.inv(mbb_array_select)

    return mbb_inv_array


def cov2corr(cov, remove_diag=True):
    """Go from covariance to correlation matrix, also setting the diagonal to zero

    Parameters
    ----------
    cov: 2d array
      the covariance matrix
    """

    d = np.sqrt(cov.diagonal())
    corr = ((cov.T/d).T)/d
    if remove_diag == True:
        corr -= np.identity(cov.shape[0])
    return corr

def selectblock(cov, spectra, n_bins, block="TTTT"):
    """Select a block in a spin0 and 2 covariance matrix

    Parameters
    ----------

    cov: 2d array
      the covariance matrix
    spectra: list of strings
      the arangement of the different block
    n_bins: int
      the number of bins for each block
    block: string
      the block you want to look at
    """

    if spectra == None:
        print ("cov mat of spin 0, no block selection needed")
        return
    else:
        blockindex = {}
        for c1,s1 in enumerate(spectra):
            for c2,s2 in enumerate(spectra):
                blockindex[s1 + s2] = [c1 * n_bins, c2 * n_bins]
    id1 = blockindex[block][0]
    id2 = blockindex[block][1]
    cov_select = cov[id1:id1 + n_bins, id2:id2 + n_bins]
    return cov_select

def delta2(a, b):
    """Simple delta function
    """

    if a == b:
        return 1
    else:
        return 0

def delta3(a, b, c):
    """Delta function (3 variables)
    """

    if (a == b) & (b == c):
        return 1
    else:
        return 0

def delta4(a, b, c, d):
    """Delta function (4 variables)
    """

    if (a == b) & (b == c) & (c == d):
        return 1
    else:
        return 0

def f(a, b, c, d, ns):
    """f combination factor in the covariance computation
    """

    result = 1. * ns[a] * (ns[c] * ns[d] * delta2(a, b) - ns[c] * delta3(a, b, d) - ns[d] * delta3(a, b, c) + delta4(a, b, c, d))
    result /= (ns[c] * ns[d] * (ns[a] - delta2(a, c)) * (ns[b] - delta2(b, d)))
    return result

def g(a, b, c, d, ns):
    """g combination factor in the covariance computation
    """

    result = 1. * ns[a] * (ns[c] * delta2(a,b) * delta2(c, d) - delta4(a, b, c, d))
    result /= (ns[a] * ns[b] * (ns[c] - delta2(a, c)) * (ns[d] - delta2(b, d)))
    return result

    
def chi(alpha, gamma, beta, eta, ns, Dl, DNl, id="TTTT"):
    """doc not ready yet
    """

    sv_alpha, ar_alpha = alpha.split("&")
    sv_beta, ar_beta = beta.split("&")
    sv_gamma, ar_gamma = gamma.split("&")
    sv_eta, ar_eta = eta.split("&")

    AB = id[0] + id[1]
    CD = id[2] + id[3]
    chi = Dl[alpha, gamma, AB] * Dl[beta, eta, CD]
    chi += Dl[alpha, gamma, AB] * DNl[beta, eta, CD] * f(sv_beta, sv_eta, sv_alpha, sv_gamma, ns)
    chi += Dl[beta, eta, CD] * DNl[alpha, gamma, AB] * f(sv_alpha, sv_gamma, sv_beta, sv_eta, ns)
    chi += g(sv_alpha, sv_gamma, sv_beta, sv_eta, ns) * DNl[alpha, gamma, AB] * DNl[beta, eta, CD]
    
    chi= symmetrize(chi, mode="arithm")

    return chi


def plot_cov_matrix(mat, color_range=None, color="pwhite", file_name=None):
    """plot the covariance matrix at full resolution using pixell plotting routines

    Parameters
    ----------

    mat: 2d array
      the covariance matrix
    color_range: float
      the range of the plot
    color: pixell colormap
      the colormap for the plot (have to be pixell compatible)
    file_name: string
      file_name is the name of the png file that will be created, if None the plot
      will be displayed.
    """

    mat_plot = mat.copy()
    try:
        colorize.mpl_setdefault(color)
    except KeyError:
        raise KeyError("Color name must be a pixell color map name {}!".format(
                list(colorize.schemes.keys())))

    if color_range is None:
        max_range = np.maximum(np.max(mat_plot),np.abs(np.min(mat_plot)))
        color_range = "%s" % (max_range)

    # We need to revert the order of the array to make the plot it similar to matplotlib imshow

    mat_plot = mat_plot[::-1, ::-1]

    wcs = enmap.create_wcs(mat_plot.shape, proj="car")
    mat_plot = enmap.enmap(mat_plot,wcs)
    plots = enplot.get_plots(mat_plot,
                             color=color,
                             range=color_range,
                             colorbar=1,
                             grid=0)
    for plot in plots:
        if file_name is not None:
            enplot.write(file_name + ".png", plot)
        else:
            plot.img.show()


def mc_cov_from_spectra_list(spec_list_a, spec_list_b, spectra=None):
    """Return the montecarlo covariance of two spectra list estimated from simulations
    This routine is inefficient and need some more work

    Parameters
    ----------
    spec_list_a: first list of spectra (or list of dict if spectra is not None)
        a list of ps computed from simulation (each entry of the list is a ps)
    spec_list_b: second list of spectra (or list of dict if spectra is not None)
        a list of ps computed from simulation (each entry of the list is a ps)

    spectra: list of strings
        needed for spin0 and spin2 cross correlation, the arrangement of the spectra
    """

    if spectra is None:
        mc_cov = np.cov(spec_list_a, y=spec_list_b,  rowvar=False)
        mean_a = np.mean(spec_list_a, axis=0)
        mean_b = np.mean(spec_list_b, axis=0)

    else:
        vec_all_a, vec_all_b = [], []
        for spec_a, spec_b in zip(spec_list_a, spec_list_b):
            vec_a, vec_b = [], []
            for spec in spectra:
                vec_a = np.append(vec_a, spec_a[spec])
                vec_b = np.append(vec_b, spec_b[spec])
            vec_all_a +=  [vec_a]
            vec_all_b +=  [vec_b]
            
        mc_cov = np.cov(vec_all_a, y=vec_all_b, rowvar=False)
        
        vec_mean_a = np.mean(vec_all_a, axis=0)
        vec_mean_b = np.mean(vec_all_b, axis=0)

        n_bins = int(len(vec_mean_a)/len(spectra))
        mean_a = {}
        mean_b = {}
        for i, spec in enumerate(spectra):
            mean_a[spec] = vec_mean_a[i * n_bins : (i + 1) * n_bins ]
            mean_b[spec] = vec_mean_b[i * n_bins : (i + 1) * n_bins ]

    n_el = int(mc_cov.shape[0] / 2)
    return mean_a, mean_b, mc_cov[n_el:, :n_el]



"""Tests for `pspy.so_map` object."""

import os
import unittest

import healpy as hp
import numpy as np
from pspy import so_map

TEST_DATA_PREFIX = os.path.join(os.path.dirname(__file__), "data")


def maps_are_equal(so_map1, so_map2):
    assert so_map1.pixel == so_map2.pixel
    assert so_map1.nside == so_map2.nside
    assert so_map1.ncomp == so_map2.ncomp
    # assert so_map1.geometry[0] == so_map2.geometry[0]
    assert so_map1.coordinate == so_map2.coordinate
    assert np.allclose(so_map1.data, so_map2.data, rtol=0, atol=1e-4)


class SOMapTests(unittest.TestCase):
    def test_box(self):
        ra0, ra1, dec0, dec1 = -1, 2, -3, 4
        box = so_map.get_box(ra0, ra1, dec0, dec1)
        box = np.rad2deg(box)
        self.assertEqual(box[0, 0], dec0)
        self.assertEqual(box[0, 1], ra1)
        self.assertEqual(box[1, 0], dec1)
        self.assertEqual(box[1, 1], ra0)

    def test_copy(self):
        ncomp, res = 3, 1
        ra0, ra1, dec0, dec1 = -1, 2, -3, 4
        so_map1 = so_map.car_template(ncomp, ra0, ra1, dec0, dec1, res)
        so_map2 = so_map1.copy()
        maps_are_equal(so_map1, so_map2)

    # def test_car_synfast(self):
    #     # Fix seed
    #     np.random.seed(14)
    #     ra0, ra1, dec0, dec1 = -5, 5, -5, 5
    #     res = 1
    #     ncomp = 3
    #     so_map1 = so_map.car_template(ncomp, ra0, ra1, dec0, dec1, res)
    #     so_map2 = so_map.read_map(os.path.join(TEST_DATA_PREFIX, "map_car.fits"))
    #     test_maps_equality(so_map1.synfast(os.path.join(TEST_DATA_PREFIX, "cl_camb.dat")), so_map2)

    # def test_healpix_synfast(self):
    #     # Fix seed
    #     np.random.seed(14)
    #     nside = 256
    #     ncomp = 3
    #     so_map1 = so_map.healpix_template(ncomp, nside)
    #     so_map2 = so_map.read_map(os.path.join(TEST_DATA_PREFIX, "map_healpix.fits"))
    #     test_maps_equality(so_map1.synfast(os.path.join(TEST_DATA_PREFIX, "cl_camb.dat")), so_map2)

    def test_subtract_mono_dipole_healpix(self):
        monopole = 0.5
        dipole = np.array([0.0, 1.0, 0.0])
        nside = 32
        m = so_map.healpix_template(ncomp=1, nside=nside)
        mask = m.copy()
        m.data = np.full_like(m.data, monopole)
        for ipix in range(hp.nside2npix(nside)):
            vec = hp.pix2vec(nside, ipix)
            m.data[ipix] += np.sum(vec * dipole)
            mask.data[ipix] = 1 if vec[0] < 0.5 else 0

        # Test raw function
        results = so_map.subtract_mono_dipole(m.data, mask.data, return_values=True)
        np.testing.assert_almost_equal(results[0], 0.0)
        self.assertAlmostEqual(results[1], monopole)
        np.testing.assert_almost_equal(results[2], dipole)

        # Test so_map method
        m.ncomp = 3
        m.data = np.tile(m.data, (3, 1))
        m.subtract_mono_dipole(mask=(mask, mask))
        np.testing.assert_almost_equal(m.data, 0.0)

    def test_subtract_mono_dipole_car(self):
        monopole = 0.5
        dipole = np.array([0.0, 1.0, 0.0])
        m = so_map.car_template(ncomp=1, res=10, ra0=-180, ra1=180, dec0=-90, dec1=90)
        mask = m.copy()
        m.data = np.full_like(m.data, monopole)

        dec, ra = m.data.posmap()
        theta = np.pi / 2 - dec
        x = np.sin(theta) * np.cos(ra)
        y = np.sin(theta) * np.sin(ra)
        z = np.cos(theta)
        mask.data = np.where(x < 0.5, 1, 0)

        m.data += np.sum(np.array([x, y, z]) * dipole[:, None, None], axis=0)

        # Test raw function
        results = so_map.subtract_mono_dipole(m.data, mask.data, healpix=False, return_values=True)
        np.testing.assert_almost_equal(results[0], 0.0)
        self.assertAlmostEqual(results[1], monopole)
        np.testing.assert_almost_equal(results[2], dipole)

        # Test so_map method
        m.ncomp = 3
        m.data = np.tile(m.data, (3, 1, 1))
        m.subtract_mono_dipole(mask=(mask, mask))
        np.testing.assert_almost_equal(m.data, 0.0)


if __name__ == "__main__":
    unittest.main()
